#!/usr/bin/env python3

import os
import shutil
import unittest
from pathlib import Path
from unittest import mock

import fdroidserver
from fdroidserver import common, signindex
from .shared_test_code import GP_FINGERPRINT, mkdtemp


basedir = Path(__file__).parent


class ApiTest(unittest.TestCase):
    """Test the public API in the base "fdroidserver" module

    This is mostly a smokecheck to make sure the public API as
    declared in fdroidserver/__init__.py is working.  The functions
    are all implemented in other modules, with their own tests.

    """

    def setUp(self):
        os.chdir(basedir)

        self._td = mkdtemp()
        self.testdir = self._td.name

        common.config = None
        config = common.read_config()
        config['jarsigner'] = common.find_sdk_tools_cmd('jarsigner')
        common.config = config
        signindex.config = config

    def tearDown(self):
        self._td.cleanup()

    def test_download_repo_index_no_fingerprint(self):
        with self.assertRaises(fdroidserver.VerificationException):
            fdroidserver.download_repo_index("http://example.org")

    @mock.patch('fdroidserver.net.http_get')
    def test_download_repo_index_url_parsing(self, mock_http_get):
        """Test whether it is trying to download the right file

        This passes the URL back via the etag return value just as a
        hack to check which URL was actually attempted.

        """
        mock_http_get.side_effect = lambda url, etag, timeout: (None, url)
        repo_url = 'https://example.org/fdroid/repo'
        index_url = 'https://example.org/fdroid/repo/index-v1.jar'
        for url in (repo_url, index_url):
            _ignored, etag_set_to_url = fdroidserver.download_repo_index(
                url, verify_fingerprint=False
            )
            self.assertEqual(index_url, etag_set_to_url)

    @mock.patch('fdroidserver.net.http_get')
    def test_download_repo_index_v1_url_parsing(self, mock_http_get):
        """Test whether it is trying to download the right file

        This passes the URL back via the etag return value just as a
        hack to check which URL was actually attempted.

        """
        mock_http_get.side_effect = lambda url, etag, timeout: (None, url)
        repo_url = 'https://example.org/fdroid/repo'
        index_url = 'https://example.org/fdroid/repo/index-v1.jar'
        for url in (repo_url, index_url):
            _ignored, etag_set_to_url = fdroidserver.download_repo_index_v1(
                url, verify_fingerprint=False
            )
            self.assertEqual(index_url, etag_set_to_url)

    @mock.patch('fdroidserver.net.download_using_mirrors')
    def test_download_repo_index_v2(self, mock_download_using_mirrors):
        """Basically a copy of IndexTest.test_download_repo_index_v2"""
        mock_download_using_mirrors.side_effect = lambda mirrors: os.path.join(
            self.testdir, 'repo', os.path.basename(mirrors[0]['url'])
        )
        os.chdir(self.testdir)
        signindex.config['keystore'] = os.path.join(basedir, 'keystore.jks')
        os.mkdir('repo')
        shutil.copy(basedir / 'repo' / 'entry.json', 'repo')
        shutil.copy(basedir / 'repo' / 'index-v2.json', 'repo')
        signindex.sign_index('repo', 'entry.json')
        repo_url = 'https://fake.url/fdroid/repo'
        entry_url = 'https://fake.url/fdroid/repo/entry.jar'
        index_url = 'https://fake.url/fdroid/repo/index-v2.json'
        fingerprint_url = 'https://fake.url/fdroid/repo?fingerprint=' + GP_FINGERPRINT
        slash_url = 'https://fake.url/fdroid/repo//?fingerprint=' + GP_FINGERPRINT
        for url in (repo_url, entry_url, index_url, fingerprint_url, slash_url):
            data, _ignored = fdroidserver.download_repo_index_v2(
                url, verify_fingerprint=False
            )
            self.assertEqual(['repo', 'packages'], list(data))
            self.assertEqual(
                'My First F-Droid Repo Demo', data['repo']['name']['en-US']
            )



#!/usr/bin/env python3
#
# metadata.py - part of the FDroid server tools
# Copyright (C) 2013, Ciaran Gultnieks, ciaran@ciarang.com
# Copyright (C) 2013-2014 Daniel Martí <mvdan@mvdan.cc>
# Copyright (C) 2017-2018 Michael Pöhn <michael.poehn@fsfe.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
import math
import platform
import os
import re
import logging
import ruamel.yaml
from collections import OrderedDict

from . import common
from . import _
from .exception import MetaDataException
from ._yaml import yaml

srclibs = None
warnings_action = None

# validates usernames based on a loose collection of rules from GitHub, GitLab,
# Liberapay and issuehunt.  This is mostly to block abuse.
VALID_USERNAME_REGEX = re.compile(r'^[a-z\d](?:[a-z\d/._-]){0,38}$', re.IGNORECASE)


def _warn_or_exception(value, cause=None):
    """Output warning or Exception depending on -W."""
    if warnings_action == 'ignore':
        pass
    elif warnings_action == 'error':
        if cause:
            raise MetaDataException(value) from cause
        else:
            raise MetaDataException(value)
    else:
        logging.warning(value)


yaml_app_field_order = [
    'Disabled',
    'AntiFeatures',
    'Categories',
    'License',
    'AuthorName',
    'AuthorEmail',
    'AuthorWebSite',
    'WebSite',
    'SourceCode',
    'IssueTracker',
    'Translation',
    'Changelog',
    'Donate',
    'Liberapay',
    'OpenCollective',
    'Bitcoin',
    'Litecoin',
    '\n',
    'Name',
    'AutoName',
    'Summary',
    'Description',
    '\n',
    'RequiresRoot',
    '\n',
    'RepoType',
    'Repo',
    'Binaries',
    '\n',
    'Builds',
    '\n',
    'AllowedAPKSigningKeys',
    '\n',
    'MaintainerNotes',
    '\n',
    'ArchivePolicy',
    'AutoUpdateMode',
    'UpdateCheckMode',
    'UpdateCheckIgnore',
    'VercodeOperation',
    'UpdateCheckName',
    'UpdateCheckData',
    'CurrentVersion',
    'CurrentVersionCode',
    '\n',
    'NoSourceSince',
]


yaml_app_fields = [x for x in yaml_app_field_order if x != '\n']


class App(dict):
    def __init__(self, copydict=None):
        if copydict:
            super().__init__(copydict)
            return
        super().__init__()

        self.Disabled = None
        self.AntiFeatures = dict()
        self.Provides = None
        self.Categories = []
        self.License = 'Unknown'
        self.AuthorName = None
        self.AuthorEmail = None
        self.AuthorWebSite = None
        self.WebSite = ''
        self.SourceCode = ''
        self.IssueTracker = ''
        self.Translation = ''
        self.Changelog = ''
        self.Donate = None
        self.Liberapay = None
        self.OpenCollective = None
        self.Bitcoin = None
        self.Litecoin = None
        self.Name = None
        self.AutoName = ''
        self.Summary = ''
        self.Description = ''
        self.RequiresRoot = False
        self.RepoType = ''
        self.Repo = ''
        self.Binaries = None
        self.AllowedAPKSigningKeys = []
        self.MaintainerNotes = ''
        self.ArchivePolicy = None
        self.AutoUpdateMode = 'None'
        self.UpdateCheckMode = 'None'
        self.UpdateCheckIgnore = None
        self.VercodeOperation = []
        self.UpdateCheckName = None
        self.UpdateCheckData = None
        self.CurrentVersion = ''
        self.CurrentVersionCode = None
        self.NoSourceSince = ''

        self.id = None
        self.metadatapath = None
        self.Builds = []
        self.added = None
        self.lastUpdated = None

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)


TYPE_STRING = 2
TYPE_BOOL = 3
TYPE_LIST = 4
TYPE_SCRIPT = 5
TYPE_MULTILINE = 6
TYPE_BUILD = 7
TYPE_INT = 8
TYPE_STRINGMAP = 9

fieldtypes = {
    'Description': TYPE_MULTILINE,
    'MaintainerNotes': TYPE_MULTILINE,
    'Categories': TYPE_LIST,
    'AntiFeatures': TYPE_STRINGMAP,
    'RequiresRoot': TYPE_BOOL,
    'AllowedAPKSigningKeys': TYPE_LIST,
    'Builds': TYPE_BUILD,
    'VercodeOperation': TYPE_LIST,
    'CurrentVersionCode': TYPE_INT,
    'ArchivePolicy': TYPE_INT,
}


def fieldtype(name):
    name = name.replace(' ', '')
    if name in fieldtypes:
        return fieldtypes[name]
    return TYPE_STRING


# In the order in which they are laid out on files
build_flags = [
    'versionName',
    'versionCode',
    'disable',
    'commit',
    'timeout',
    'subdir',
    'submodules',
    'sudo',
    'init',
    'patch',
    'gradle',
    'maven',
    'output',
    'binary',
    'srclibs',
    'oldsdkloc',
    'encoding',
    'forceversion',
    'forcevercode',
    'rm',
    'extlibs',
    'prebuild',
    'androidupdate',
    'target',
    'scanignore',
    'scandelete',
    'build',
    'buildjni',
    'ndk',
    'preassemble',
    'gradleprops',
    'antcommands',
    'postbuild',
    'novcheck',
    'antifeatures',
]


class Build(dict):
    def __init__(self, copydict=None):
        super().__init__()
        self.disable = ''
        self.commit = None
        self.timeout = None
        self.subdir = None
        self.submodules = False
        self.sudo = ''
        self.init = ''
        self.patch = []
        self.gradle = []
        self.maven = None
        self.output = None
        self.binary = None
        self.srclibs = []
        self.oldsdkloc = False
        self.encoding = None
        self.forceversion = False
        self.forcevercode = False
        self.rm = []
        self.extlibs = []
        self.prebuild = ''
        self.androidupdate = []
        self.target = None
        self.scanignore = []
        self.scandelete = []
        self.build = ''
        self.buildjni = []
        self.ndk = None
        self.preassemble = []
        self.gradleprops = []
        self.antcommands = []
        self.postbuild = ''
        self.novcheck = False
        self.antifeatures = dict()
        if copydict:
            super().__init__(copydict)
            return

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node)

    def build_method(self):
        for f in ['maven', 'gradle']:
            if self.get(f):
                return f
        if self.output:
            return 'raw'
        return 'ant'

    # like build_method, but prioritize output=
    def output_method(self):
        if self.output:
            return 'raw'
        for f in ['maven', 'gradle']:
            if self.get(f):
                return f
        return 'ant'

    def ndk_path(self) -> str:
        """Return the path string of the first configured NDK or an empty string."""
        ndk = self.ndk
        if isinstance(ndk, list):
            ndk = self.ndk[0]
        path = common.config['ndk_paths'].get(ndk)
        if path and not isinstance(path, str):
            raise TypeError('NDK path is not string')
        if path:
            return path
        for vsn, path in common.config['ndk_paths'].items():
            if not vsn.endswith("_orig") and path and os.path.basename(path) == ndk:
                return path
        return ''


flagtypes = {
    'versionCode': TYPE_INT,
    'extlibs': TYPE_LIST,
    'srclibs': TYPE_LIST,
    'patch': TYPE_LIST,
    'rm': TYPE_LIST,
    'buildjni': TYPE_LIST,
    'preassemble': TYPE_LIST,
    'androidupdate': TYPE_LIST,
    'scanignore': TYPE_LIST,
    'scandelete': TYPE_LIST,
    'gradle': TYPE_LIST,
    'antcommands': TYPE_LIST,
    'gradleprops': TYPE_LIST,
    'sudo': TYPE_SCRIPT,
    'init': TYPE_SCRIPT,
    'prebuild': TYPE_SCRIPT,
    'build': TYPE_SCRIPT,
    'postbuild': TYPE_SCRIPT,
    'submodules': TYPE_BOOL,
    'oldsdkloc': TYPE_BOOL,
    'forceversion': TYPE_BOOL,
    'forcevercode': TYPE_BOOL,
    'novcheck': TYPE_BOOL,
    'antifeatures': TYPE_STRINGMAP,
    'timeout': TYPE_INT,
}


def flagtype(name):
    if name in flagtypes:
        return flagtypes[name]
    return TYPE_STRING


class FieldValidator:
    """Designate App metadata field types and checks that it matches.

    'name'     - The long name of the field type
    'matching' - List of possible values or regex expression
    'sep'      - Separator to use if value may be a list
    'fields'   - Metadata fields (Field:Value) of this type
    """

    def __init__(self, name, matching, fields):
        self.name = name
        self.matching = matching
        self.compiled = re.compile(matching)
        self.fields = fields

    def check(self, v, appid):
        if not v:
            return
        if type(v) == list:
            values = v
        else:
            values = [v]
        for v in values:
            if not self.compiled.match(v):
                _warn_or_exception(
                    _(
                        "'{value}' is not a valid {field} in {appid}. Regex pattern: {pattern}"
                    ).format(
                        value=v, field=self.name, appid=appid, pattern=self.matching
                    )
                )


# Generic value types
valuetypes = {
    FieldValidator("Liberapay",
                   VALID_USERNAME_REGEX,
                   ['Liberapay']),

    FieldValidator("Open Collective",
                   VALID_USERNAME_REGEX,
                   ['OpenCollective']),

    FieldValidator("HTTP link",
                   r'^http[s]?://',
                   ["WebSite", "SourceCode", "IssueTracker", "Translation", "Changelog", "Donate"]),

    FieldValidator("Email",
                   r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$',
                   ["AuthorEmail"]),

    FieldValidator("Bitcoin address",
                   r'^(bc1|[13])[a-zA-HJ-NP-Z0-9]{25,39}$',
                   ["Bitcoin"]),

    FieldValidator("Litecoin address",
                   r'^([LM3][a-km-zA-HJ-NP-Z1-9]{26,33}|ltc1[a-z0-9]{39})$',
                   ["Litecoin"]),

    FieldValidator("Repo Type",
                   r'^(git|git-svn|svn|hg|bzr|srclib)$',
                   ["RepoType"]),

    FieldValidator("Binaries",
                   r'^http[s]?://',
                   ["Binaries"]),

    FieldValidator("AllowedAPKSigningKeys",
                   r'^[a-fA-F0-9]{64}$',
                   ["AllowedAPKSigningKeys"]),

    FieldValidator("Auto Update Mode",
                   r"^(Version.*|None)$",
                   ["AutoUpdateMode"]),

    FieldValidator("Update Check Mode",
                   r"^(Tags|Tags .+|RepoManifest|RepoManifest/.+|HTTP|Static|None)$",
                   ["UpdateCheckMode"])
}


# Check an app's metadata information for integrity errors
def check_metadata(app):
    for v in valuetypes:
        for k in v.fields:
            v.check(app[k], app.id)


def parse_yaml_srclib(metadatapath):
    thisinfo = {'RepoType': '', 'Repo': '', 'Subdir': None, 'Prepare': None}

    if not metadatapath.exists():
        _warn_or_exception(
            _("Invalid scrlib metadata: '{file}' does not exist").format(
                file=metadatapath
            )
        )
        return thisinfo

    with metadatapath.open("r", encoding="utf-8") as f:
        try:
            data = yaml.load(f)
            if type(data) is not dict:
                if platform.system() == 'Windows':
                    # Handle symlink on Windows
                    symlink = metadatapath.parent / metadatapath.read_text(encoding='utf-8')
                    if symlink.is_file():
                        with symlink.open("r", encoding="utf-8") as s:
                            data = yaml.load(s)
            if type(data) is not dict:
                raise ruamel.yaml.YAMLError(
                    _('{file} is blank or corrupt!').format(file=metadatapath)
                )
        except ruamel.yaml.YAMLError as e:
            _warn_or_exception(_("Invalid srclib metadata: could not "
                                 "parse '{file}'")
                               .format(file=metadatapath) + '\n'
                               + common.run_yamllint(metadatapath, indent=4),
                               cause=e)
            return thisinfo

    for key in data:
        if key not in thisinfo:
            _warn_or_exception(
                _("Invalid srclib metadata: unknown key '{key}' in '{file}'").format(
                    key=key, file=metadatapath
                )
            )
            return thisinfo
        else:
            if key == 'Subdir':
                if isinstance(data[key], str):
                    thisinfo[key] = data[key].split(',')
                elif isinstance(data[key], list):
                    thisinfo[key] = data[key]
                elif data[key] is None:
                    thisinfo[key] = ['']
            elif key == 'Prepare' or flagtype(key) == TYPE_SCRIPT:
                if isinstance(data[key], list):
                    thisinfo[key] = data[key]
                else:
                    thisinfo[key] = [data[key]] if data[key] else []
            else:
                thisinfo[key] = str(data[key] or '')

    return thisinfo


def read_srclibs():
    """Read all srclib metadata.

    The information read will be accessible as metadata.srclibs, which is a
    dictionary, keyed on srclib name, with the values each being a dictionary
    in the same format as that returned by the parse_yaml_srclib function.

    A MetaDataException is raised if there are any problems with the srclib
    metadata.
    """
    global srclibs

    # They were already loaded
    if srclibs is not None:
        return

    srclibs = {}

    srclibs_dir = Path('srclibs')
    srclibs_dir.mkdir(exist_ok=True)

    for metadatapath in sorted(srclibs_dir.glob('*.yml')):
        srclibs[metadatapath.stem] = parse_yaml_srclib(metadatapath)


def read_metadata(appid_to_vercode={}, sort_by_time=False):
    """Return a list of App instances sorted newest first.

    This reads all of the metadata files in a 'data' repository, then
    builds a list of App instances from those files.  The list is
    sorted based on creation time, newest first.  Most of the time,
    the newer files are the most interesting.

    appid_to_vercode is a dict with appids a keys and versionCodes as values.

    """
    # Always read the srclibs before the apps, since they can use a srlib as
    # their source repository.
    read_srclibs()

    apps = OrderedDict()

    for basedir in ('metadata', 'tmp'):
        Path(basedir).mkdir(exist_ok=True)

    if appid_to_vercode:
        metadatafiles = common.get_metadata_files(appid_to_vercode)
    else:
        metadatafiles = list(Path('metadata').glob('*.yml')) + list(
            Path('.').glob('.fdroid.yml')
        )

    if sort_by_time:
        entries = ((path.stat().st_mtime, path) for path in metadatafiles)
        metadatafiles = []
        for _ignored, path in sorted(entries, reverse=True):
            metadatafiles.append(path)
    else:
        # most things want the index alpha sorted for stability
        metadatafiles = sorted(metadatafiles)

    for metadatapath in metadatafiles:
        appid = metadatapath.stem
        if appid != '.fdroid' and not common.is_valid_package_name(appid):
            _warn_or_exception(
                _("{appid} from {path} is not a valid Java Package Name!").format(
                    appid=appid, path=metadatapath
                )
            )
        if appid in apps:
            _warn_or_exception(
                _("Found multiple metadata files for {appid}").format(appid=appid)
            )
        app = parse_metadata(metadatapath)
        check_metadata(app)
        apps[app.id] = app

    return apps


def parse_metadata(metadatapath):
    """Parse metadata file, also checking the source repo for .fdroid.yml.

    This function finds the relevant files, gets them parsed, converts
    dicts into App and Build instances, and combines the results into
    a single App instance.

    If this is a metadata file from fdroiddata, it will first load the
    source repo type and URL from fdroiddata, then read .fdroid.yml if
    it exists, then include the rest of the metadata as specified in
    fdroiddata, so that fdroiddata has precedence over the metadata in
    the source code.

    .fdroid.yml is embedded in the app's source repo, so it is
    "user-generated".  That means that it can have weird things in it
    that need to be removed so they don't break the overall process,
    e.g. if the upstream developer includes some broken field, it can
    be overridden in the metadata file.

    Parameters
    ----------
    metadatapath
      The file path to read. The "Application ID" aka "Package Name"
      for the application comes from this filename.

    Raises
    ------
    FDroidException when there are syntax errors.

    Returns
    -------
    Returns a dictionary containing all the details of the
    application. There are two major kinds of information in the
    dictionary. Keys beginning with capital letters correspond
    directory to identically named keys in the metadata file.  Keys
    beginning with lower case letters are generated in one way or
    another, and are not found verbatim in the metadata.

    """
    metadatapath = Path(metadatapath)
    app = App()
    app.metadatapath = metadatapath.as_posix()
    if metadatapath.suffix == '.yml':
        with metadatapath.open('r', encoding='utf-8') as mf:
            app.update(parse_yaml_metadata(mf))
    else:
        _warn_or_exception(
            _('Unknown metadata format: {path} (use: *.yml)').format(path=metadatapath)
        )

    if metadatapath.stem != '.fdroid':
        app.id = metadatapath.stem
        parse_localized_antifeatures(app)

    if metadatapath.name != '.fdroid.yml' and app.Repo:
        build_dir = common.get_build_dir(app)
        metadata_in_repo = build_dir / '.fdroid.yml'
        if metadata_in_repo.is_file():
            commit_id = common.get_head_commit_id(build_dir)
            if commit_id is not None:
                logging.debug(
                    _('Including metadata from %s@%s') % (metadata_in_repo, commit_id)
                )
            else:
                logging.debug(
                    _('Including metadata from {path}').format(path=metadata_in_repo)
                )
            app_in_repo = parse_metadata(metadata_in_repo)
            for k, v in app_in_repo.items():
                if k not in app:
                    app[k] = v

    builds = []
    for build in app.get('Builds', []):
        builds.append(Build(build))
    if builds:
        app['Builds'] = builds

    # if only .fdroid.yml was found, then this finds the appid
    if not app.id:
        if app.get('Builds'):
            build = app['Builds'][-1]
            if build.subdir:
                root_dir = Path(build.subdir)
            else:
                root_dir = Path('.')
            paths = common.manifest_paths(root_dir, build.gradle)
            _ignored, _ignored, app.id = common.parse_androidmanifests(paths, app)

    return app


def parse_yaml_metadata(mf):
    """Parse the .yml file and post-process it.

    This function handles parsing a metadata YAML file and converting
    all the various data types into a consistent internal
    representation.  The results are meant to update an existing App
    instance or used as a plain dict.

    Clean metadata .yml files can be used directly, but in order to
    make a better user experience for people editing .yml files, there
    is post processing.  That makes the parsing perform something like
    Strict YAML.

    """
    try:
        yamldata = common.yaml.load(mf)
    except ruamel.yaml.YAMLError as e:
        _warn_or_exception(
            _("could not parse '{path}'").format(path=mf.name)
            + '\n'
            + common.run_yamllint(mf.name, indent=4),
            cause=e,
        )

    if yamldata is None or yamldata == '':
        yamldata = dict()
    if not isinstance(yamldata, dict):
        _warn_or_exception(
            _("'{path}' has invalid format, it should be a dictionary!").format(
                path=mf.name
            )
        )
        logging.error(_('Using blank dictionary instead of contents of {path}!').format(
            path=mf.name)
        )
        yamldata = dict()

    deprecated_in_yaml = ['Provides']

    for field in tuple(yamldata.keys()):
        if field not in yaml_app_fields + deprecated_in_yaml:
            msg = _("Unrecognised app field '{fieldname}' in '{path}'").format(
                fieldname=field, path=mf.name
            )
            if Path(mf.name).name == '.fdroid.yml':
                logging.error(msg)
                del yamldata[field]
            else:
                _warn_or_exception(msg)

    for deprecated_field in deprecated_in_yaml:
        if deprecated_field in yamldata:
            del yamldata[deprecated_field]
            logging.warning(
                _(
                    "Ignoring '{field}' in '{metapath}' "
                    "metadata because it is deprecated."
                ).format(field=deprecated_field, metapath=mf.name)
            )

    msg = _("Unrecognised build flag '{build_flag}' in '{path}'")
    for build in yamldata.get('Builds', []):
        for build_flag in build:
            if build_flag not in build_flags:
                _warn_or_exception(msg.format(build_flag=build_flag, path=mf.name))

    post_parse_yaml_metadata(yamldata)
    return yamldata


def parse_localized_antifeatures(app):
    """Read in localized Anti-Features files from the filesystem.

    To support easy integration with Weblate and other translation
    systems, there is a special type of metadata that can be
    maintained in a Fastlane-style directory layout, where each field
    is represented by a text file on directories that specified which
    app it belongs to, which locale, etc.  This function reads those
    in and puts them into the internal dict, to be merged with any
    related data that came from the metadata.yml file.

    This needs to be run after parse_yaml_metadata() since that
    normalizes the data structure.  Also, these values are lower
    priority than what comes from the metadata file. So this should
    not overwrite anything parse_yaml_metadata() puts into the App
    instance.

    metadata/<Application ID>/<locale>/antifeatures/<Version Code>_<Anti-Feature>.txt
    metadata/<Application ID>/<locale>/antifeatures/<Anti-Feature>.txt

    └── metadata/
        └── <Application ID>/
            ├── en-US/
            │   └── antifeatures/
            │       ├── 123_Ads.txt       -> "includes ad lib"
            │       ├── 123_Tracking.txt  -> "standard suspects"
            │       └── NoSourceSince.txt -> "it vanished"
            │
            └── zh-CN/
                └── antifeatures/
                    └── 123_Ads.txt       -> "包括广告库"

    Gets parsed into the metadata data structure:

    AntiFeatures:
      NoSourceSince:
        en-US: it vanished
    Builds:
      - versionCode: 123
        antifeatures:
          Ads:
            en-US: includes ad lib
            zh-CN: 包括广告库
          Tracking:
            en-US: standard suspects

    """
    app_dir = Path('metadata', app['id'])
    if not app_dir.is_dir():
        return
    af_dup_msg = _('Duplicate Anti-Feature declaration at {path} was ignored!')

    if app.get('AntiFeatures'):
        app_has_AntiFeatures = True
    else:
        app_has_AntiFeatures = False

    has_versionCode = re.compile(r'^-?[0-9]+_.*')
    has_antifeatures_from_app = set()
    for build in app.get('Builds', []):
        antifeatures = build.get('antifeatures')
        if antifeatures:
            has_antifeatures_from_app.add(build['versionCode'])

    for f in sorted(app_dir.glob('*/antifeatures/*.txt')):
        path = f.as_posix()
        left = path.index('/', 9)  # 9 is length of "metadata/"
        right = path.index('/', left + 1)
        locale = path[left + 1 : right]
        description = f.read_text()
        if has_versionCode.match(f.stem):
            i = f.stem.index('_')
            versionCode = int(f.stem[:i])
            antifeature = f.stem[i + 1 :]
            if versionCode in has_antifeatures_from_app:
                logging.error(af_dup_msg.format(path=f))
                continue
            if 'Builds' not in app:
                app['Builds'] = []
            found = False
            for build in app['Builds']:
                # loop though builds again, there might be duplicate versionCodes
                if versionCode == build['versionCode']:
                    found = True
                    if 'antifeatures' not in build:
                        build['antifeatures'] = dict()
                    if antifeature not in build['antifeatures']:
                        build['antifeatures'][antifeature] = dict()
                    build['antifeatures'][antifeature][locale] = description
            if not found:
                app['Builds'].append(
                    {
                        'versionCode': versionCode,
                        'antifeatures': {
                            antifeature: {locale: description},
                        },
                    }
                )
        elif app_has_AntiFeatures:
            logging.error(af_dup_msg.format(path=f))
            continue
        else:
            if 'AntiFeatures' not in app:
                app['AntiFeatures'] = dict()
            if f.stem not in app['AntiFeatures']:
                app['AntiFeatures'][f.stem] = dict()
            app['AntiFeatures'][f.stem][locale] = f.read_text()


def _normalize_type_int(k, v):
    """Normalize anything that can be reliably converted to an integer."""
    if isinstance(v, int) and not isinstance(v, bool):
        return v
    if v is None:
        return None
    if isinstance(v, str):
        try:
            return int(v)
        except ValueError:
            pass
    msg = _('{build_flag} must be an integer, found: {value}')
    _warn_or_exception(msg.format(build_flag=k, value=v))


def _normalize_type_string(v):
    """Normalize any data to TYPE_STRING.

    YAML 1.2's booleans are all lowercase.

    Things like versionName are strings, but without quotes can be
    numbers.  Like "versionName: 1.0" would be a YAML float, but
    should be a string.

    SHA-256 values are string values, but YAML 1.2 can interpret some
    unquoted values as decimal ints.  This converts those to a string
    if they are over 50 digits.  In the wild, the longest 0 padding on
    a SHA-256 key fingerprint I found was 8 zeros.

    """
    if isinstance(v, bool):
        if v:
            return 'true'
        return 'false'
    if isinstance(v, float):
        # YAML 1.2 values for NaN, Inf, and -Inf
        if math.isnan(v):
            return '.nan'
        if math.isinf(v):
            if v > 0:
                return '.inf'
            return '-.inf'
    if v and isinstance(v, int):
        if math.log10(v) > 50:  # only if the int has this many digits
            return '%064d' % v
    return str(v)


def _normalize_type_stringmap(k, v):
    """Normalize any data to TYPE_STRINGMAP.

    The internal representation of this format is a dict of dicts,
    where the outer dict's keys are things like tag names of
    Anti-Features, the inner dict's keys are locales, and the ultimate
    values are human readable text.

    Metadata entries like AntiFeatures: can be written in many
    forms, including a simple one-entry string, a list of strings,
    a dict with keys and descriptions as values, or a dict with
    localization.

    Returns
    -------
    A dictionary with string keys, where each value is either a string
    message or a dict with locale keys and string message values.

    """
    if v is None:
        return dict()
    if isinstance(v, str) or isinstance(v, int) or isinstance(v, float):
        return {_normalize_type_string(v): dict()}
    if isinstance(v, list) or isinstance(v, tuple) or isinstance(v, set):
        retdict = dict()
        for i in v:
            if isinstance(i, dict):
                # transitional format
                if len(i) != 1:
                    _warn_or_exception(
                        _(
                            "'{value}' is not a valid {field}, should be {pattern}"
                        ).format(field=k, value=v, pattern='key: value')
                    )
                afname = _normalize_type_string(next(iter(i)))
                desc = _normalize_type_string(next(iter(i.values())))
                retdict[afname] = {common.DEFAULT_LOCALE: desc}
            else:
                retdict[_normalize_type_string(i)] = {}
        return retdict

    retdict = dict()
    for af, afdict in v.items():
        key = _normalize_type_string(af)
        if afdict:
            if isinstance(afdict, dict):
                retdict[key] = afdict
            else:
                retdict[key] = {common.DEFAULT_LOCALE: _normalize_type_string(afdict)}
        else:
            retdict[key] = dict()

    return retdict


def _normalize_type_list(k, v):
    """Normalize any data to TYPE_LIST, which is always a list of strings."""
    if isinstance(v, dict):
        msg = _('{build_flag} must be list or string, found: {value}')
        _warn_or_exception(msg.format(build_flag=k, value=v))
    elif type(v) not in (list, tuple, set):
        v = [v]
    return [_normalize_type_string(i) for i in v]


def post_parse_yaml_metadata(yamldata):
    """Convert human-readable metadata data structures into consistent data structures.

    "Be conservative in what is written out, be liberal in what is parsed."
    https://en.wikipedia.org/wiki/Robustness_principle

    This also handles conversions that make metadata YAML behave
    something like StrictYAML.  Specifically, a field should have a
    fixed value type, regardless of YAML 1.2's type auto-detection.

    TODO: None values should probably be treated as the string 'null',
    since YAML 1.2 uses that for nulls

    """
    for k, v in yamldata.items():
        _fieldtype = fieldtype(k)
        if _fieldtype == TYPE_LIST:
            if v or v == 0:
                yamldata[k] = _normalize_type_list(k, v)
        elif _fieldtype == TYPE_INT:
            # ArchivePolicy used to require " versions" in the value.
            if k == 'ArchivePolicy' and isinstance(v, str):
                v = v.split(' ', maxsplit=1)[0]
            v = _normalize_type_int(k, v)
            if v or v == 0:
                yamldata[k] = v
        elif _fieldtype == TYPE_STRING:
            if v or v == 0:
                yamldata[k] = _normalize_type_string(v)
        elif _fieldtype == TYPE_STRINGMAP:
            if v or v == 0:  # TODO probably want just `if v:`
                yamldata[k] = _normalize_type_stringmap(k, v)
        elif _fieldtype == TYPE_BOOL:
            yamldata[k] = bool(v)
        else:
            if type(v) in (float, int):
                yamldata[k] = str(v)

    builds = []
    for build in yamldata.get('Builds', []):
        for k, v in build.items():
            if v is None:
                continue

            _flagtype = flagtype(k)
            if _flagtype == TYPE_STRING:
                if v or v == 0:
                    build[k] = _normalize_type_string(v)
            elif _flagtype == TYPE_INT:
                v = _normalize_type_int(k, v)
                if v or v == 0:
                    build[k] = v
            elif _flagtype in (TYPE_LIST, TYPE_SCRIPT):
                if v or v == 0:
                    build[k] = _normalize_type_list(k, v)
            elif _flagtype == TYPE_STRINGMAP:
                if v or v == 0:
                    build[k] = _normalize_type_stringmap(k, v)
            elif _flagtype == TYPE_BOOL:
                build[k] = bool(v)

        builds.append(build)

    if builds:
        yamldata['Builds'] = sorted(builds, key=lambda build: build['versionCode'])

    no_source_since = yamldata.get("NoSourceSince")
    # do not overwrite the description if it is there
    if no_source_since and not yamldata.get('AntiFeatures', {}).get('NoSourceSince'):
        if 'AntiFeatures' not in yamldata:
            yamldata['AntiFeatures'] = dict()
        yamldata['AntiFeatures']['NoSourceSince'] = {
            common.DEFAULT_LOCALE: no_source_since
        }


def _format_multiline(value):
    """TYPE_MULTILINE with newlines in them are saved as YAML literal strings."""
    if '\n' in value:
        return ruamel.yaml.scalarstring.preserve_literal(str(value))
    return str(value)


def _format_list(value):
    """TYPE_LIST should not contain null values."""
    return [v for v in value if v]


def _format_script(value):
    """TYPE_SCRIPT with one value are converted to YAML string values."""
    value = [v for v in value if v]
    if len(value) == 1:
        return value[0]
    return value


def _format_stringmap(appid, field, stringmap, versionCode=None):
    """Format TYPE_STRINGMAP taking into account localized files in the metadata dir.

    If there are any localized versions on the filesystem already,
    then move them all there.  Otherwise, keep them in the .yml file.

    The directory for the localized files that is named after the
    field is all lower case, following the convention set by Fastlane
    metadata, and used by fdroidserver.

    """
    app_dir = Path('metadata', appid)
    try:
        next(app_dir.glob('*/%s/*.txt' % field.lower()))
        files = []
        overwrites = []
        for name, descdict in stringmap.items():
            for locale, desc in descdict.items():
                outdir = app_dir / locale / field.lower()
                if versionCode:
                    filename = '%d_%s.txt' % (versionCode, name)
                else:
                    filename = '%s.txt' % name
                outfile = outdir / filename
                files.append(str(outfile))
                if outfile.exists():
                    if desc != outfile.read_text():
                        overwrites.append(str(outfile))
                else:
                    if not outfile.parent.exists():
                        outfile.parent.mkdir(parents=True)
                    outfile.write_text(desc)
        if overwrites:
            _warn_or_exception(
                _(
                    'Conflicting "{field}" definitions between .yml and localized files:'
                ).format(field=field)
                + '\n'
                + '\n'.join(sorted(overwrites))
            )
        logging.warning(
            _('Moving Anti-Features declarations to localized files:')
            + '\n'
            + '\n'.join(sorted(files))
        )
        return
    except StopIteration:
        pass
    make_list = True
    outlist = []
    for name in sorted(stringmap):
        outlist.append(name)
        descdict = stringmap.get(name)
        if descdict and any(descdict.values()):
            make_list = False
            break
    if make_list:
        return sorted(outlist, key=str.lower)
    return stringmap


def _del_duplicated_NoSourceSince(app):
    # noqa: D403 NoSourceSince is the word.
    """NoSourceSince gets auto-added to AntiFeatures, but can also be manually added."""
    key = 'NoSourceSince'
    if key in app:
        no_source_since = app.get(key)
        af_no_source_since = app.get('AntiFeatures', dict()).get(key)
        if af_no_source_since == {common.DEFAULT_LOCALE: no_source_since}:
            del app['AntiFeatures'][key]


def _builds_to_yaml(app):
    """Reformat Builds: flags for output to YAML 1.2.

    This will strip any flag/value that is not set or is empty.
    TYPE_BOOL fields are removed when they are false.  0 is valid
    value, it should not be stripped, so there are special cases to
    handle that.

    """
    builds = ruamel.yaml.comments.CommentedSeq()
    for build in app.get('Builds', []):
        b = ruamel.yaml.comments.CommentedMap()
        for field in build_flags:
            v = build.get(field)
            if v is None or v is False or v == '' or v == dict() or v == list():
                continue
            _flagtype = flagtype(field)
            if _flagtype == TYPE_MULTILINE:
                v = _format_multiline(v)
            elif _flagtype == TYPE_LIST:
                v = _format_list(v)
            elif _flagtype == TYPE_SCRIPT:
                v = _format_script(v)
            elif _flagtype == TYPE_STRINGMAP:
                v = _format_stringmap(app['id'], field, v, build['versionCode'])

            if v or v == 0:
                b[field] = v

        builds.append(b)

    # insert extra empty lines between build entries
    for i in range(1, len(builds)):
        builds.yaml_set_comment_before_after_key(i, 'bogus')
        builds.ca.items[i][1][-1].value = '\n'

    return builds


def _app_to_yaml(app):
    cm = ruamel.yaml.comments.CommentedMap()
    insert_newline = False
    for field in yaml_app_field_order:
        if field == '\n':
            # next iteration will need to insert a newline
            insert_newline = True
        else:
            value = app.get(field)
            if value or field in ('Builds', 'ArchivePolicy'):
                _fieldtype = fieldtype(field)
                if field == 'Builds':
                    if app.get('Builds'):
                        cm.update({field: _builds_to_yaml(app)})
                elif field == 'Categories':
                    cm[field] = sorted(value, key=str.lower)
                elif field == 'AntiFeatures':
                    v = _format_stringmap(app['id'], field, value)
                    if v:
                        cm[field] = v
                elif field == 'AllowedAPKSigningKeys':
                    value = [str(i).lower() for i in value]
                    if len(value) == 1:
                        cm[field] = value[0]
                    else:
                        cm[field] = value
                elif field == 'ArchivePolicy':
                    if value is None:
                        continue
                    cm[field] = value
                elif _fieldtype == TYPE_MULTILINE:
                    v = _format_multiline(value)
                    if v:
                        cm[field] = v
                elif _fieldtype == TYPE_SCRIPT:
                    v = _format_script(value)
                    if v:
                        cm[field] = v
                else:
                    if value:
                        cm[field] = value

                if insert_newline:
                    # we need to prepend a newline in front of this field
                    insert_newline = False
                    # inserting empty lines is not supported so we add a
                    # bogus comment and over-write its value
                    cm.yaml_set_comment_before_after_key(field, 'bogus')
                    cm.ca.items[field][1][-1].value = '\n'
    return cm


def write_yaml(mf, app):
    """Write metadata in yaml format.

    This requires the 'rt' round trip dumper to maintain order and needs
    custom indent settings, so it needs to instantiate its own YAML
    instance.  Therefore, this function deliberately avoids using any of
    the common YAML parser setups.

    Parameters
    ----------
    mf
      active file discriptor for writing
    app
      app metadata to written to the YAML file

    """
    _del_duplicated_NoSourceSince(app)
    yaml_app = _app_to_yaml(app)
    yamlmf = ruamel.yaml.YAML(typ='rt')
    yamlmf.indent(mapping=2, sequence=4, offset=2)
    yamlmf.dump(yaml_app, stream=mf)


def write_metadata(metadatapath, app):
    metadatapath = Path(metadatapath)
    if metadatapath.suffix == '.yml':
        with metadatapath.open('w') as mf:
            return write_yaml(mf, app)

    _warn_or_exception(_('Unknown metadata format: %s') % metadatapath)


def add_metadata_arguments(parser):
    """Add common command line flags related to metadata processing."""
    parser.add_argument(
        "-W",
        choices=['error', 'warn', 'ignore'],
        default='error',
        help=_("force metadata errors (default) to be warnings, or to be ignored."),
    )



import itertools
import os
import platform
import re
import shlex
import shutil
import subprocess
import sys
import threading
import unittest
from datetime import datetime, timezone
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path

try:
    from androguard.core.bytecodes.apk import get_apkid  # androguard <4
except ModuleNotFoundError:
    from androguard.core.apk import get_apkid

from fdroidserver._yaml import yaml, yaml_dumper
from .shared_test_code import mkdir_testfiles

# TODO: port generic tests that use index.xml to index-v2 (test that
#       explicitly test index-v0 should still use index.xml)


basedir = Path(__file__).parent
FILES = basedir

try:
    WORKSPACE = Path(os.environ["WORKSPACE"])
except KeyError:
    WORKSPACE = basedir.parent

from fdroidserver import common

conf = {"sdk_path": os.getenv("ANDROID_HOME", "")}
common.find_apksigner(conf)
USE_APKSIGNER = "apksigner" in conf


@unittest.skipIf(sys.byteorder == 'big', 'androguard is not ported to big-endian')
class IntegrationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        try:
            cls.fdroid_cmd = shlex.split(os.environ["fdroid"])
        except KeyError:
            cls.fdroid_cmd = [WORKSPACE / "fdroid"]

        os.environ.update(
            {
                "GIT_AUTHOR_NAME": "Test",
                "GIT_AUTHOR_EMAIL": "no@mail",
                "GIT_COMMITTER_NAME": "Test",
                "GIT_COMMITTER_EMAIL": "no@mail",
                "GIT_ALLOW_PROTOCOL": "file:https",
            }
        )

    def setUp(self):
        self.prev_cwd = Path()
        self.testdir = mkdir_testfiles(WORKSPACE, self)
        self.tmp_repo_root = self.testdir / "fdroid"
        self.tmp_repo_root.mkdir(parents=True)
        os.chdir(self.tmp_repo_root)

    def tearDown(self):
        os.chdir(self.prev_cwd)
        shutil.rmtree(self.testdir)

    def assert_run(self, *args, **kwargs):
        proc = subprocess.run(*args, **kwargs)
        self.assertEqual(proc.returncode, 0)
        return proc

    def assert_run_fail(self, *args, **kwargs):
        proc = subprocess.run(*args, **kwargs)
        self.assertNotEqual(proc.returncode, 0)
        return proc

    @staticmethod
    def update_yaml(path, items, replace=False):
        """Update a .yml file, e.g. config.yml, with the given items."""
        doc = {}
        if not replace:
            try:
                with open(path) as f:
                    doc = yaml.load(f)
            except FileNotFoundError:
                pass
        doc.update(items)
        with open(path, "w") as f:
            yaml_dumper.dump(doc, f)

    @staticmethod
    def remove_lines(path, unwanted_strings):
        """Remove the lines in the path that contain the unwanted strings."""

        def contains_unwanted(line, unwanted_strings):
            for str in unwanted_strings:
                if str in line:
                    return True
            return False

        with open(path) as f:
            filtered = [
                line for line in f if not contains_unwanted(line, unwanted_strings)
            ]

        with open(path, "w") as f:
            for line in filtered:
                f.write(line)

    @staticmethod
    def copy_apks_into_repo():
        def to_skip(name):
            for str in [
                "unaligned",
                "unsigned",
                "badsig",
                "badcert",
                "bad-unicode",
                "janus.apk",
            ]:
                if str in name:
                    return True
            return False

        for f in FILES.glob("*.apk"):
            if not to_skip(f.name):
                appid, versionCode, _ignored = get_apkid(f)
                shutil.copy(
                    f,
                    Path("repo") / common.get_release_apk_filename(appid, versionCode),
                )

    @staticmethod
    def create_fake_android_home(path):
        (path / "tools").mkdir()
        (path / "platform-tools").mkdir()
        (path / "build-tools/34.0.0").mkdir(parents=True)
        (path / "build-tools/34.0.0/aapt").touch()

    def fdroid_init_with_prebuilt_keystore(self, keystore_path=FILES / "keystore.jks"):
        self.assert_run(
            self.fdroid_cmd
            + ["init", "--keystore", keystore_path, "--repo-keyalias", "sova"]
        )
        self.update_yaml(
            common.CONFIG_FILE,
            {
                "keystorepass": "r9aquRHYoI8+dYz6jKrLntQ5/NJNASFBacJh7Jv2BlI=",
                "keypass": "r9aquRHYoI8+dYz6jKrLntQ5/NJNASFBacJh7Jv2BlI=",
            },
        )

    @unittest.skipUnless(USE_APKSIGNER, "requires apksigner")
    def test_run_process_when_building_and_signing_are_on_separate_machines(self):
        shutil.copy(FILES / "keystore.jks", "keystore.jks")
        self.fdroid_init_with_prebuilt_keystore("keystore.jks")
        self.update_yaml(
            common.CONFIG_FILE,
            {
                "make_current_version_link": True,
                "keydname": "CN=Birdman, OU=Cell, O=Alcatraz, L=Alcatraz, S=California, C=US",
            },
        )

        Path("metadata").mkdir()
        shutil.copy(FILES / "metadata/info.guardianproject.urzip.yml", "metadata")
        Path("unsigned").mkdir()
        shutil.copy(
            FILES / "urzip-release-unsigned.apk",
            "unsigned/info.guardianproject.urzip_100.apk",
        )

        self.assert_run(self.fdroid_cmd + ["publish", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["update", "--verbose", "--nosign"])
        self.assert_run(self.fdroid_cmd + ["signindex", "--verbose"])

        self.assertIn(
            '<application id="info.guardianproject.urzip">',
            Path("repo/index.xml").read_text(),
        )
        self.assertTrue(Path("repo/index.jar").is_file())
        self.assertTrue(Path("repo/index-v1.jar").is_file())
        apkcache = Path("tmp/apkcache.json")
        self.assertTrue(apkcache.is_file())
        self.assertTrue(apkcache.stat().st_size > 0)
        self.assertTrue(Path("urzip.apk").is_symlink())

    def test_utf8_metadata(self):
        self.fdroid_init_with_prebuilt_keystore()
        self.update_yaml(
            common.CONFIG_FILE,
            {
                "repo_description": "获取已安装在您的设备上的应用的",
                "mirrors": ["https://foo.bar/fdroid", "http://secret.onion/fdroid"],
            },
        )
        shutil.copy(FILES / "urzip.apk", "repo")
        shutil.copy(FILES / "bad-unicode-πÇÇ现代通用字-български-عربي1.apk", "repo")
        Path("metadata").mkdir()
        shutil.copy(FILES / "metadata/info.guardianproject.urzip.yml", "metadata")

        self.assert_run(self.fdroid_cmd + ["readmeta"])
        self.assert_run(self.fdroid_cmd + ["update"])

    def test_copy_git_import_and_run_fdroid_scanner_on_it(self):
        url = "https://gitlab.com/fdroid/ci-test-app.git"
        Path("metadata").mkdir()
        self.update_yaml(
            "metadata/org.fdroid.ci.test.app.yml",
            {
                "AutoName": "Just A Test",
                "WebSite": None,
                "Builds": [
                    {
                        "versionName": "0.3",
                        "versionCode": 300,
                        "commit": "0.3",
                        "subdir": "app",
                        "gradle": ["yes"],
                    }
                ],
                "Repo": url,
                "RepoType": "git",
            },
        )

        self.assert_run(["git", "clone", url, "build/org.fdroid.ci.test.app"])
        self.assert_run(
            self.fdroid_cmd + ["scanner", "org.fdroid.ci.test.app", "--verbose"]
        )

    @unittest.skipUnless(shutil.which("gpg"), "requires command line gpg")
    def test_copy_repo_generate_java_gpg_keys_update_and_gpgsign(self):
        """Needs tricks to make gpg-agent run in a test harness."""
        self.fdroid_init_with_prebuilt_keystore()
        shutil.copytree(FILES / "repo", "repo", dirs_exist_ok=True)
        for dir in ["config", "metadata"]:
            shutil.copytree(FILES / dir, dir)
        # gpg requires a short path to the socket to talk to gpg-agent
        gnupghome = (WORKSPACE / '.testfiles/gnupghome').resolve()
        shutil.rmtree(gnupghome, ignore_errors=True)
        shutil.copytree(FILES / "gnupghome", gnupghome)
        os.chmod(gnupghome, 0o700)
        self.update_yaml(
            common.CONFIG_FILE,
            {
                "install_list": "org.adaway",
                "uninstall_list": ["com.android.vending", "com.facebook.orca"],
                "gpghome": str(gnupghome),
                "gpgkey": "CE71F7FB",
                "mirrors": [
                    "http://foobarfoobarfoobar.onion/fdroid",
                    "https://foo.bar/fdroid",
                ],
            },
        )
        self.assert_run(
            self.fdroid_cmd + ["update", "--verbose", "--pretty"],
            env=os.environ | {"LC_MESSAGES": "C.UTF-8"},
        )
        index_xml = Path("repo/index.xml").read_text()
        self.assertIn("<application id=", index_xml)
        self.assertIn("<install packageName=", index_xml)
        self.assertIn("<uninstall packageName=", index_xml)
        self.assertTrue(Path("repo/index.jar").is_file())
        self.assertTrue(Path("repo/index-v1.jar").is_file())

        self.assert_run(self.fdroid_cmd + ["gpgsign", "--verbose"])
        env = os.environ.copy()
        env["GNUPGHOME"] = gnupghome
        self.assert_run(['gpgconf', '--kill', 'gpg-agent'], env=env)

        self.assertTrue(Path("repo/obb.mainpatch.current_1619.apk.asc").is_file())
        self.assertTrue(
            Path("repo/obb.main.twoversions_1101617_src.tar.gz.asc").is_file()
        )
        self.assertFalse(Path("repo/obb.mainpatch.current_1619.apk.asc.asc").exists())
        self.assertFalse(
            Path("repo/obb.main.twoversions_1101617_src.tar.gz.asc.asc").exists()
        )
        self.assertFalse(Path("repo/index.xml.asc").exists())

        index_v1_json = Path("repo/index-v1.json").read_text()
        v0_timestamp = re.search(r'timestamp="(\d+)"', index_xml).group(1)
        v1_timestamp = re.search(r'"timestamp": (\d+)', index_v1_json).group(1)[:-3]
        self.assertEqual(v0_timestamp, v1_timestamp)

        # we can't easily reproduce the timestamps for things, so just hardcode them
        index_xml = re.sub(r'timestamp="\d+"', 'timestamp="1676634233"', index_xml)
        self.assertEqual((FILES / "repo/index.xml").read_text(), index_xml)
        index_v1_json = re.sub(
            r'"timestamp": (\d+)', '"timestamp": 1676634233000', index_v1_json
        )
        self.assertEqual((FILES / "repo/index-v1.json").read_text(), index_v1_json)

        expected_index_v2_json = (FILES / "repo/index-v2.json").read_text()
        expected_index_v2_json = re.sub(
            r',\s*"ipfsCIDv1":\s*"[\w]+"', "", expected_index_v2_json
        )

        index_v2_json = Path("repo/index-v2.json").read_text()
        index_v2_json = re.sub(
            r'"timestamp": (\d+)', '"timestamp": 1676634233000', index_v2_json
        )
        index_v2_json = re.sub(r',\s*"ipfsCIDv1":\s*"[\w]+"', "", index_v2_json)
        self.assertEqual(index_v2_json, expected_index_v2_json)

    def test_moving_lots_of_apks_to_the_archive(self):
        self.fdroid_init_with_prebuilt_keystore()
        Path("metadata").mkdir()
        for path in (FILES / "metadata").glob("*.yml"):
            shutil.copy(path, "metadata")
        self.update_yaml(
            "metadata/info.guardianproject.urzip.yml",
            {"Summary": "good test version of urzip"},
            replace=True,
        )
        self.update_yaml(
            "metadata/org.bitbucket.tickytacky.mirrormirror.yml",
            {"Summary": "good MD5 sig, which is disabled algorithm"},
            replace=True,
        )
        for f in Path("metadata").glob("*.yml"):
            self.remove_lines(f, ["ArchivePolicy:"])
        for f in itertools.chain(
            FILES.glob("urzip.apk"),
            FILES.glob("org.bitbucket.tickytacky.mirrormirror_[0-9].apk"),
            FILES.glob("repo/com.politedroid_[0-9].apk"),
            FILES.glob("repo/obb.main.twoversions_110161[357].apk"),
        ):
            shutil.copy(f, "repo")
        self.update_yaml(common.CONFIG_FILE, {"archive_older": 3})

        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        with open("archive/index.xml") as f:
            archive_cnt = sum(1 for line in f if "<package>" in line)
        with open("repo/index.xml") as f:
            repo_cnt = sum(1 for line in f if "<package>" in line)
        if USE_APKSIGNER:
            self.assertEqual(archive_cnt, 2)
            self.assertEqual(repo_cnt, 10)
        else:
            # This will fail when jarsigner allows MD5 for APK signatures
            self.assertEqual(archive_cnt, 5)
            self.assertEqual(repo_cnt, 7)

    @unittest.skipIf(USE_APKSIGNER, "runs only without apksigner")
    def test_per_app_archive_policy(self):
        self.fdroid_init_with_prebuilt_keystore()
        Path("metadata").mkdir()
        shutil.copy(FILES / "metadata/com.politedroid.yml", "metadata")
        for f in FILES.glob("repo/com.politedroid_[0-9].apk"):
            shutil.copy(f, "repo")
        self.update_yaml(common.CONFIG_FILE, {"archive_older": 3})

        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 4)
        self.assertEqual(archive_cnt, 0)
        self.assertIn("com.politedroid_3.apk", repo)
        self.assertIn("com.politedroid_4.apk", repo)
        self.assertIn("com.politedroid_5.apk", repo)
        self.assertIn("com.politedroid_6.apk", repo)
        self.assertTrue(Path("repo/com.politedroid_3.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_4.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_5.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())

        # enable one app in the repo
        self.update_yaml("metadata/com.politedroid.yml", {"ArchivePolicy": 1})
        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 1)
        self.assertEqual(archive_cnt, 3)
        self.assertIn("com.politedroid_6.apk", repo)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertIn("com.politedroid_4.apk", archive)
        self.assertIn("com.politedroid_5.apk", archive)
        self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_4.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_5.apk").is_file())

        # remove all apps from the repo
        self.update_yaml("metadata/com.politedroid.yml", {"ArchivePolicy": 0})
        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 0)
        self.assertEqual(archive_cnt, 4)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertIn("com.politedroid_4.apk", archive)
        self.assertIn("com.politedroid_5.apk", archive)
        self.assertIn("com.politedroid_6.apk", archive)
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_4.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_5.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_6.apk").is_file())
        self.assertFalse(Path("repo/com.politedroid_6.apk").exists())

        # move back one from archive to the repo
        self.update_yaml("metadata/com.politedroid.yml", {"ArchivePolicy": 1})
        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 1)
        self.assertEqual(archive_cnt, 3)
        self.assertIn("com.politedroid_6.apk", repo)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertIn("com.politedroid_4.apk", archive)
        self.assertIn("com.politedroid_5.apk", archive)
        self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_4.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_5.apk").is_file())
        self.assertFalse(Path("archive/com.politedroid_6.apk").exists())

        # set an earlier version as CVC and test that it's the only one not archived
        self.update_yaml("metadata/com.politedroid.yml", {"CurrentVersionCode": 5})
        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 1)
        self.assertEqual(archive_cnt, 3)
        self.assertIn("com.politedroid_5.apk", repo)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertIn("com.politedroid_4.apk", archive)
        self.assertIn("com.politedroid_6.apk", archive)
        self.assertTrue(Path("repo/com.politedroid_5.apk").is_file())
        self.assertFalse(Path("repo/com.politedroid_6.apk").exists())
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_4.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_6.apk").is_file())

    def test_moving_old_apks_to_and_from_the_archive(self):
        self.fdroid_init_with_prebuilt_keystore()
        Path("metadata").mkdir()
        shutil.copy(FILES / "metadata/com.politedroid.yml", "metadata")
        self.remove_lines("metadata/com.politedroid.yml", ["ArchivePolicy:"])
        for f in FILES.glob("repo/com.politedroid_[0-9].apk"):
            shutil.copy(f, "repo")
        self.update_yaml(common.CONFIG_FILE, {"archive_older": 3})

        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 3)
        self.assertIn("com.politedroid_4.apk", repo)
        self.assertIn("com.politedroid_5.apk", repo)
        self.assertIn("com.politedroid_6.apk", repo)
        self.assertTrue(Path("repo/com.politedroid_4.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_5.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(archive_cnt, 1)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())

        self.update_yaml(common.CONFIG_FILE, {"archive_older": 1})
        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 1)
        self.assertIn("com.politedroid_6.apk", repo)
        self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(archive_cnt, 3)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertIn("com.politedroid_4.apk", archive)
        self.assertIn("com.politedroid_5.apk", archive)
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_4.apk").is_file())
        self.assertTrue(Path("archive/com.politedroid_5.apk").is_file())

        # disabling deletes from the archive
        metadata_path = Path("metadata/com.politedroid.yml")
        metadata = metadata_path.read_text()
        metadata = re.sub(
            "versionCode: 4", "versionCode: 4\n    disable: testing deletion", metadata
        )
        metadata_path.write_text(metadata)
        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 1)
        self.assertIn("com.politedroid_6.apk", repo)
        self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(archive_cnt, 2)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertNotIn("com.politedroid_4.apk", archive)
        self.assertIn("com.politedroid_5.apk", archive)
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertFalse(Path("archive/com.politedroid_4.apk").exists())
        self.assertTrue(Path("archive/com.politedroid_5.apk").is_file())

        # disabling deletes from the repo, and promotes one from the archive
        metadata = re.sub(
            "versionCode: 6", "versionCode: 6\n    disable: testing deletion", metadata
        )
        metadata_path.write_text(metadata)
        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 1)
        self.assertIn("com.politedroid_5.apk", repo)
        self.assertNotIn("com.politedroid_6.apk", repo)
        self.assertTrue(Path("repo/com.politedroid_5.apk").is_file())
        self.assertFalse(Path("repo/com.politedroid_6.apk").exists())
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(archive_cnt, 1)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertFalse(Path("archive/com.politedroid_6.apk").exists())

    def test_that_verify_can_succeed_and_fail(self):
        Path("tmp").mkdir()
        Path("unsigned").mkdir()
        shutil.copy(FILES / "repo/com.politedroid_6.apk", "tmp")
        shutil.copy(FILES / "repo/com.politedroid_6.apk", "unsigned")
        self.assert_run(
            self.fdroid_cmd
            + ["verify", "--reuse-remote-apk", "--verbose", "com.politedroid"]
        )
        # force a fail
        shutil.copy(
            FILES / "repo/com.politedroid_5.apk", "unsigned/com.politedroid_6.apk"
        )
        self.assert_run_fail(
            self.fdroid_cmd
            + ["verify", "--reuse-remote-apk", "--verbose", "com.politedroid"]
        )

    def test_allowing_disabled_signatures_in_repo_and_archive(self):
        self.fdroid_init_with_prebuilt_keystore()
        self.update_yaml(
            common.CONFIG_FILE, {"allow_disabled_algorithms": True, "archive_older": 3}
        )
        Path("metadata").mkdir()
        shutil.copy(FILES / "metadata/com.politedroid.yml", "metadata")
        self.update_yaml(
            "metadata/info.guardianproject.urzip.yml",
            {"Summary": "good test version of urzip"},
            replace=True,
        )
        self.update_yaml(
            "metadata/org.bitbucket.tickytacky.mirrormirror.yml",
            {"Summary": "good MD5 sig, disabled algorithm"},
            replace=True,
        )
        for f in Path("metadata").glob("*.yml"):
            self.remove_lines(f, ["ArchivePolicy:"])
        for f in itertools.chain(
            FILES.glob("urzip-badsig.apk"),
            FILES.glob("org.bitbucket.tickytacky.mirrormirror_[0-9].apk"),
            FILES.glob("repo/com.politedroid_[0-9].apk"),
        ):
            shutil.copy(f, "repo")

        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        repo = Path("repo/index.xml").read_text()
        repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
        archive = Path("archive/index.xml").read_text()
        archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
        self.assertEqual(repo_cnt, 6)
        self.assertEqual(archive_cnt, 2)
        self.assertIn("com.politedroid_4.apk", repo)
        self.assertIn("com.politedroid_5.apk", repo)
        self.assertIn("com.politedroid_6.apk", repo)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertIn("org.bitbucket.tickytacky.mirrormirror_2.apk", repo)
        self.assertIn("org.bitbucket.tickytacky.mirrormirror_3.apk", repo)
        self.assertIn("org.bitbucket.tickytacky.mirrormirror_4.apk", repo)
        self.assertIn("org.bitbucket.tickytacky.mirrormirror_1.apk", archive)
        self.assertNotIn("urzip-badsig.apk", repo)
        self.assertNotIn("urzip-badsig.apk", archive)
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_4.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_5.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())
        self.assertTrue(
            Path("archive/org.bitbucket.tickytacky.mirrormirror_1.apk").is_file()
        )
        self.assertTrue(
            Path("repo/org.bitbucket.tickytacky.mirrormirror_2.apk").is_file()
        )
        self.assertTrue(
            Path("repo/org.bitbucket.tickytacky.mirrormirror_3.apk").is_file()
        )
        self.assertTrue(
            Path("repo/org.bitbucket.tickytacky.mirrormirror_4.apk").is_file()
        )
        self.assertTrue(Path("archive/urzip-badsig.apk").is_file())

        if not USE_APKSIGNER:
            self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
            repo = Path("repo/index.xml").read_text()
            repo_cnt = sum(1 for line in repo.splitlines() if "<package>" in line)
            archive = Path("archive/index.xml").read_text()
            archive_cnt = sum(1 for line in archive.splitlines() if "<package>" in line)
            self.assertEqual(repo_cnt, 3)
            self.assertEqual(archive_cnt, 5)
            self.assertIn("com.politedroid_4.apk", repo)
            self.assertIn("com.politedroid_5.apk", repo)
            self.assertIn("com.politedroid_6.apk", repo)
            self.assertNotIn("urzip-badsig.apk", repo)
            self.assertIn("org.bitbucket.tickytacky.mirrormirror_1.apk", archive)
            self.assertIn("org.bitbucket.tickytacky.mirrormirror_2.apk", archive)
            self.assertIn("org.bitbucket.tickytacky.mirrormirror_3.apk", archive)
            self.assertIn("org.bitbucket.tickytacky.mirrormirror_4.apk", archive)
            self.assertIn("com.politedroid_3.apk", archive)
            self.assertNotIn("urzip-badsig.apk", archive)
            self.assertTrue(Path("repo/com.politedroid_4.apk").is_file())
            self.assertTrue(Path("repo/com.politedroid_5.apk").is_file())
            self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())
            self.assertTrue(
                Path("archive/org.bitbucket.tickytacky.mirrormirror_1.apk").is_file()
            )
            self.assertTrue(
                Path("archive/org.bitbucket.tickytacky.mirrormirror_2.apk").is_file()
            )
            self.assertTrue(
                Path("archive/org.bitbucket.tickytacky.mirrormirror_3.apk").is_file()
            )
            self.assertTrue(
                Path("archive/org.bitbucket.tickytacky.mirrormirror_4.apk").is_file()
            )
            self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
            self.assertTrue(Path("archive/urzip-badsig.apk").is_file())

        # test unarchiving when disabled_algorithms are allowed again
        self.update_yaml(common.CONFIG_FILE, {"allow_disabled_algorithms": True})
        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        with open("archive/index.xml") as f:
            archive_cnt = sum(1 for line in f if "<package>" in line)
        with open("repo/index.xml") as f:
            repo_cnt = sum(1 for line in f if "<package>" in line)
        self.assertEqual(repo_cnt, 6)
        self.assertEqual(archive_cnt, 2)
        self.assertIn("com.politedroid_4.apk", repo)
        self.assertIn("com.politedroid_5.apk", repo)
        self.assertIn("com.politedroid_6.apk", repo)
        self.assertIn("org.bitbucket.tickytacky.mirrormirror_2.apk", repo)
        self.assertIn("org.bitbucket.tickytacky.mirrormirror_3.apk", repo)
        self.assertIn("org.bitbucket.tickytacky.mirrormirror_4.apk", repo)
        self.assertNotIn("urzip-badsig.apk", repo)
        self.assertIn("com.politedroid_3.apk", archive)
        self.assertIn("org.bitbucket.tickytacky.mirrormirror_1.apk", archive)
        self.assertNotIn("urzip-badsig.apk", archive)
        self.assertTrue(Path("repo/com.politedroid_4.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_5.apk").is_file())
        self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())
        self.assertTrue(
            Path("repo/org.bitbucket.tickytacky.mirrormirror_2.apk").is_file()
        )
        self.assertTrue(
            Path("repo/org.bitbucket.tickytacky.mirrormirror_3.apk").is_file()
        )
        self.assertTrue(
            Path("repo/org.bitbucket.tickytacky.mirrormirror_4.apk").is_file()
        )
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertTrue(
            Path("archive/org.bitbucket.tickytacky.mirrormirror_1.apk").is_file()
        )
        self.assertTrue(Path("archive/urzip-badsig.apk").is_file())

    def test_rename_apks_with_fdroid_update_rename_apks_opt_nosign_opt_for_speed(self):
        self.fdroid_init_with_prebuilt_keystore()
        self.update_yaml(
            common.CONFIG_FILE,
            {
                "keydname": "CN=Birdman, OU=Cell, O=Alcatraz, L=Alcatraz, S=California, C=US"
            },
        )
        Path("metadata").mkdir()
        shutil.copy(FILES / "metadata/info.guardianproject.urzip.yml", "metadata")
        shutil.copy(
            FILES / "urzip.apk",
            "repo/asdfiuhk urzip-πÇÇπÇÇ现代汉语通用字-български-عربي1234 ö.apk",
        )
        self.assert_run(
            self.fdroid_cmd + ["update", "--rename-apks", "--pretty", "--nosign"]
        )
        self.assertTrue(Path("repo/info.guardianproject.urzip_100.apk").is_file())
        index_xml = Path("repo/index.xml").read_text()
        index_v1_json = Path("repo/index-v1.json").read_text()
        self.assertIn("info.guardianproject.urzip_100.apk", index_v1_json)
        self.assertIn("info.guardianproject.urzip_100.apk", index_xml)

        shutil.copy(FILES / "urzip-release.apk", "repo")
        self.assert_run(
            self.fdroid_cmd + ["update", "--rename-apks", "--pretty", "--nosign"]
        )
        self.assertTrue(Path("repo/info.guardianproject.urzip_100.apk").is_file())
        self.assertTrue(
            Path("repo/info.guardianproject.urzip_100_b4964fd.apk").is_file()
        )
        index_xml = Path("repo/index.xml").read_text()
        index_v1_json = Path("repo/index-v1.json").read_text()
        self.assertIn("info.guardianproject.urzip_100.apk", index_v1_json)
        self.assertIn("info.guardianproject.urzip_100.apk", index_xml)
        self.assertIn("info.guardianproject.urzip_100_b4964fd.apk", index_v1_json)
        self.assertNotIn("info.guardianproject.urzip_100_b4964fd.apk", index_xml)

        shutil.copy(FILES / "urzip-release.apk", "repo")
        self.assert_run(
            self.fdroid_cmd + ["update", "--rename-apks", "--pretty", "--nosign"]
        )
        self.assertTrue(Path("repo/info.guardianproject.urzip_100.apk").is_file())
        self.assertTrue(
            Path("repo/info.guardianproject.urzip_100_b4964fd.apk").is_file()
        )
        self.assertTrue(
            Path("duplicates/repo/info.guardianproject.urzip_100_b4964fd.apk").is_file()
        )
        index_xml = Path("repo/index.xml").read_text()
        index_v1_json = Path("repo/index-v1.json").read_text()
        self.assertIn("info.guardianproject.urzip_100.apk", index_v1_json)
        self.assertIn("info.guardianproject.urzip_100.apk", index_xml)
        self.assertIn("info.guardianproject.urzip_100_b4964fd.apk", index_v1_json)
        self.assertNotIn("info.guardianproject.urzip_100_b4964fd.apk", index_xml)

    def test_for_added_date_being_set_correctly_for_repo_and_archive(self):
        self.fdroid_init_with_prebuilt_keystore()
        self.update_yaml(common.CONFIG_FILE, {"archive_older": 3})
        Path("metadata").mkdir()
        Path("archive").mkdir()
        shutil.copy(FILES / "repo/com.politedroid_6.apk", "repo")
        shutil.copy(FILES / "repo/index-v2.json", "repo")
        shutil.copy(FILES / "repo/com.politedroid_5.apk", "archive")
        shutil.copy(FILES / "metadata/com.politedroid.yml", "metadata")

        # TODO: the timestamp of the oldest apk in the file should be used, even
        #       if that doesn't exist anymore
        self.update_yaml("metadata/com.politedroid.yml", {"ArchivePolicy": 1})

        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        timestamp = int(datetime(2017, 6, 23, tzinfo=timezone.utc).timestamp()) * 1000
        index_v1_json = Path("repo/index-v1.json").read_text()
        self.assertIn(f'"added": {timestamp}', index_v1_json)
        # the archive will have the added timestamp for the app and for the apk,
        # both need to be there
        with open("archive/index-v1.json") as f:
            count = sum(1 for line in f if f'"added": {timestamp}' in line)
        self.assertEqual(count, 2)

    def test_whatsnew_from_fastlane_without_cvc_set(self):
        self.fdroid_init_with_prebuilt_keystore()
        Path("metadata/com.politedroid/en-US/changelogs").mkdir(parents=True)
        shutil.copy(FILES / "repo/com.politedroid_6.apk", "repo")
        shutil.copy(FILES / "metadata/com.politedroid.yml", "metadata")
        self.remove_lines("metadata/com.politedroid.yml", ["CurrentVersion:"])
        Path("metadata/com.politedroid/en-US/changelogs/6.txt").write_text(
            "whatsnew test"
        )
        self.assert_run(self.fdroid_cmd + ["update", "--pretty", "--nosign"])
        index_v1_json = Path("repo/index-v1.json").read_text()
        self.assertIn("whatsnew test", index_v1_json)

    def test_metadata_checks(self):
        Path("repo").mkdir()
        shutil.copy(FILES / "urzip.apk", "repo")
        # this should fail because there is no metadata
        self.assert_run_fail(self.fdroid_cmd + ["build"])
        Path("metadata").mkdir()
        shutil.copy(FILES / "metadata/org.smssecure.smssecure.yml", "metadata")
        self.assert_run(self.fdroid_cmd + ["readmeta"])

    def test_ensure_commands_that_dont_need_the_jdk_work_without_a_jdk_configured(self):
        Path("repo").mkdir()
        Path("metadata").mkdir()
        self.update_yaml(
            "metadata/fake.yml",
            {
                "License": "GPL-2.0-only",
                "Summary": "Yup still fake",
                "Categories": ["Internet"],
                "Description": "this is fake",
            },
        )
        # fake that no JDKs are available
        self.update_yaml(
            common.CONFIG_FILE,
            {"categories": ["Internet"], "java_paths": {}},
            replace=True,
        )
        local_copy_dir = self.testdir / "local_copy_dir/fdroid"
        (local_copy_dir / "repo").mkdir(parents=True)
        self.update_yaml(
            common.CONFIG_FILE, {"local_copy_dir": str(local_copy_dir.resolve())}
        )

        subprocess.run(self.fdroid_cmd + ["checkupdates", "--allow-dirty"])
        if shutil.which("gpg"):
            self.assert_run(self.fdroid_cmd + ["gpgsign"])
        self.assert_run(self.fdroid_cmd + ["lint"])
        self.assert_run(self.fdroid_cmd + ["readmeta"])
        self.assert_run(self.fdroid_cmd + ["rewritemeta", "fake"])
        self.assert_run(self.fdroid_cmd + ["deploy"])
        self.assert_run(self.fdroid_cmd + ["scanner"])

        # run these to get their output, but the are not setup, so don't fail
        subprocess.run(self.fdroid_cmd + ["build"])
        subprocess.run(self.fdroid_cmd + ["import"])
        subprocess.run(self.fdroid_cmd + ["install", "-n"])

    def test_config_checks_of_local_copy_dir(self):
        self.assert_run(self.fdroid_cmd + ["init"])
        self.assert_run(self.fdroid_cmd + ["update", "--create-metadata", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["readmeta"])
        local_copy_dir = (self.testdir / "local_copy_dir/fdroid").resolve()
        local_copy_dir.mkdir(parents=True)
        self.assert_run(
            self.fdroid_cmd + ["deploy", "--local-copy-dir", local_copy_dir]
        )
        self.assert_run(
            self.fdroid_cmd
            + ["deploy", "--local-copy-dir", local_copy_dir, "--verbose"]
        )

        # this should fail because thisisnotanabsolutepath is not an absolute path
        self.assert_run_fail(
            self.fdroid_cmd + ["deploy", "--local-copy-dir", "thisisnotanabsolutepath"]
        )
        # this should fail because the path doesn't end with "fdroid"
        self.assert_run_fail(
            self.fdroid_cmd
            + [
                "deploy",
                "--local-copy-dir",
                "/tmp/IReallyDoubtThisPathExistsasdfasdf",  # nosec B108
            ]
        )
        # this should fail because the dirname path does not exist
        self.assert_run_fail(
            self.fdroid_cmd
            + [
                "deploy",
                "--local-copy-dir",
                "/tmp/IReallyDoubtThisPathExistsasdfasdf/fdroid",  # nosec B108
            ]
        )

    def test_setup_a_new_repo_from_scratch_using_android_home_and_do_a_local_sync(self):
        self.fdroid_init_with_prebuilt_keystore()
        self.copy_apks_into_repo()
        self.assert_run(self.fdroid_cmd + ["update", "--create-metadata", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["readmeta"])
        self.assertIn("<application id=", Path("repo/index.xml").read_text())

        local_copy_dir = self.testdir / "local_copy_dir/fdroid"
        local_copy_dir.parent.mkdir()
        self.assert_run(
            self.fdroid_cmd + ["deploy", "--local-copy-dir", local_copy_dir]
        )

        new_tmp_repo = self.testdir / "new_repo"
        new_tmp_repo.mkdir()
        os.chdir(new_tmp_repo)
        self.fdroid_init_with_prebuilt_keystore()
        self.update_yaml(common.CONFIG_FILE, {"sync_from_local_copy_dir": True})
        self.assert_run(
            self.fdroid_cmd + ["deploy", "--local-copy-dir", local_copy_dir]
        )

    def test_check_that_android_home_opt_fails_when_dir_does_not_exist_or_is_not_a_dir(
        self,
    ):
        # this should fail because /opt/fakeandroidhome does not exist
        self.assert_run_fail(
            self.fdroid_cmd
            + [
                "init",
                "--keystore",
                "keystore.p12",
                "--android-home",
                "/opt/fakeandroidhome",
            ]
        )
        Path("test_file").touch()
        # this should fail because test_file is not a directory
        self.assert_run_fail(
            self.fdroid_cmd
            + ["init", "--keystore", "keystore.p12", "--android-home", "test_file"]
        )

    def test_check_that_fake_android_home_passes_fdroid_init(self):
        android_home = self.testdir / "android-sdk"
        android_home.mkdir()
        self.create_fake_android_home(android_home)
        self.assert_run(
            self.fdroid_cmd
            + ["init", "--keystore", "keystore.p12", "--android-home", android_home]
        )

    @unittest.skip
    def test_check_that_fdroid_init_fails_when_build_tools_cannot_be_found(self):
        fake_android_home = self.testdir / "android-sdk"
        fake_android_home.mkdir()
        self.create_fake_android_home(fake_android_home)
        (fake_android_home / "build-tools/34.0.0/aapt").unlink()
        self.assert_run_fail(
            self.fdroid_cmd
            + [
                "init",
                "--keystore",
                "keystore.p12",
                "--android-home",
                fake_android_home,
            ]
        )

    def check_that_android_home_opt_overrides_android_home_env_var(self):
        fake_android_home = self.testdir / "android-sdk"
        fake_android_home.mkdir()
        self.create_fake_android_home(fake_android_home)
        self.assert_run(
            self.fdroid_cmd
            + [
                "init",
                "--keystore",
                "keystore.p12",
                "--android-home",
                fake_android_home,
            ]
        )
        # the value set in --android-home should override $ANDROID_HOME
        self.assertIn(str(fake_android_home), Path(common.CONFIG_FILE).read_text())

    @unittest.skipUnless(
        "ANDROID_HOME" in os.environ, "runs only with ANDROID_HOME set"
    )
    def setup_a_new_repo_from_scratch_with_keystore_and_android_home_opt_set_on_cmd_line(
        self,
    ):
        """Test with broken setup in ANDROID_HOME.

        In this case, ANDROID_HOME is set to a fake, non-working
        version that will be detected by fdroid as an Android SDK
        install. It should use the path set by --android-home over the
        one in ANDROID_HOME, therefore if it uses the one in
        ANDROID_HOME, it won't work because it is a fake one.  Only
        --android-home provides a working one.

        """
        real_android_home = os.environ["ANDROID_HOME"]
        fake_android_home = self.testdir / "android-sdk"
        fake_android_home.mkdir()
        env = os.environ.copy()
        env["ANDROID_HOME"] = str(fake_android_home)
        self.assert_run(
            self.fdroid_cmd
            + [
                "init",
                "--keystore",
                "keystore.p12",
                "--android-home",
                real_android_home,
                "--no-prompt",
            ],
            env=env,
        )
        self.assertTrue(Path("keystore.p12").is_file())
        self.copy_apks_into_repo()
        self.assert_run(
            self.fdroid_cmd + ["update", "--create-metadata", "--verbose"], env=env
        )
        self.assert_run(self.fdroid_cmd + ["readmeta"], env=env)
        self.assertIn("<application id=", Path("repo/index.xml").read_text())
        self.assertTrue(Path("repo/index.jar").is_file())
        self.assertTrue(Path("repo/index-v1.jar").is_file())
        apkcache = Path("tmp/apkcache.json")
        self.assertTrue(apkcache.is_file())
        self.assertTrue(apkcache.stat().st_size > 0)

    def test_check_duplicate_files_are_properly_handled_by_fdroid_update(self):
        self.fdroid_init_with_prebuilt_keystore()
        Path("metadata").mkdir()
        shutil.copy(FILES / "metadata/obb.mainpatch.current.yml", "metadata")
        shutil.copy(FILES / "repo/obb.mainpatch.current_1619.apk", "repo")
        shutil.copy(
            FILES / "repo/obb.mainpatch.current_1619_another-release-key.apk", "repo"
        )
        self.assert_run(self.fdroid_cmd + ["update", "--pretty"])
        index_xml = Path("repo/index.xml").read_text()
        index_v1_json = Path("repo/index-v1.json").read_text()
        self.assertNotIn(
            "obb.mainpatch.current_1619_another-release-key.apk", index_xml
        )
        self.assertIn("obb.mainpatch.current_1619.apk", index_xml)
        self.assertIn("obb.mainpatch.current_1619.apk", index_v1_json)
        self.assertIn(
            "obb.mainpatch.current_1619_another-release-key.apk", index_v1_json
        )
        # die if there are exact duplicates
        shutil.copy(FILES / "repo/obb.mainpatch.current_1619.apk", "repo/duplicate.apk")
        self.assert_run_fail(self.fdroid_cmd + ["update"])

    def test_setup_new_repo_from_scratch_using_android_home_env_var_putting_apks_in_repo_first(
        self,
    ):
        Path("repo").mkdir()
        self.copy_apks_into_repo()
        self.fdroid_init_with_prebuilt_keystore()
        self.assert_run(self.fdroid_cmd + ["update", "--create-metadata", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["readmeta"])
        self.assertIn("<application id=", Path("repo/index.xml").read_text())

    def test_setup_a_new_repo_from_scratch_and_generate_a_keystore(self):
        self.assert_run(self.fdroid_cmd + ["init", "--keystore", "keystore.p12"])
        self.assertTrue(Path("keystore.p12").is_file())
        self.copy_apks_into_repo()
        self.assert_run(self.fdroid_cmd + ["update", "--create-metadata", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["readmeta"])
        self.assertIn("<application id=", Path("repo/index.xml").read_text())
        self.assertTrue(Path("repo/index.jar").is_file())
        self.assertTrue(Path("repo/index-v1.jar").is_file())
        apkcache = Path("tmp/apkcache.json")
        self.assertTrue(apkcache.is_file())
        self.assertTrue(apkcache.stat().st_size > 0)

    def test_setup_a_new_repo_manually_and_generate_a_keystore(self):
        self.assertFalse(Path("keystore.p12").exists())
        # this should fail because this repo has no keystore
        self.assert_run_fail(self.fdroid_cmd + ["update"])
        self.assert_run(self.fdroid_cmd + ["update", "--create-key"])
        self.assertTrue(Path("keystore.p12").is_file())
        self.copy_apks_into_repo()
        self.assert_run(self.fdroid_cmd + ["update", "--create-metadata", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["readmeta"])
        self.assertIn("<application id=", Path("repo/index.xml").read_text())
        self.assertTrue(Path("repo/index.jar").is_file())
        self.assertTrue(Path("repo/index-v1.jar").is_file())
        apkcache = Path("tmp/apkcache.json")
        self.assertTrue(apkcache.is_file())
        self.assertTrue(apkcache.stat().st_size > 0)

    def test_setup_a_new_repo_from_scratch_generate_a_keystore_then_add_apk_and_update(
        self,
    ):
        self.assert_run(self.fdroid_cmd + ["init", "--keystore", "keystore.p12"])
        self.assertTrue(Path("keystore.p12").is_file())
        self.copy_apks_into_repo()
        self.assert_run(self.fdroid_cmd + ["update", "--create-metadata", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["readmeta"])
        self.assertIn("<application id=", Path("repo/index.xml").read_text())
        self.assertTrue(Path("repo/index.jar").is_file())
        self.assertTrue(Path("repo/index-v1.jar").is_file())

        if not Path("repo/info.guardianproject.urzip_100.apk").exists():
            shutil.copy(FILES / "urzip.apk", "repo")
        self.assert_run(self.fdroid_cmd + ["update", "--create-metadata", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["readmeta"])
        self.assertIn("<application id=", Path("repo/index.xml").read_text())
        self.assertTrue(Path("repo/index.jar").is_file())
        self.assertTrue(Path("repo/index-v1.jar").is_file())
        apkcache = Path("tmp/apkcache.json")
        self.assertTrue(apkcache.is_file())
        self.assertTrue(apkcache.stat().st_size > 0)
        self.assertIn("<application id=", Path("repo/index.xml").read_text())

    def test_setup_a_new_repo_from_scratch_with_a_hsm_or_smartcard(self):
        self.assert_run(self.fdroid_cmd + ["init", "--keystore", "NONE"])
        self.assertTrue(Path("opensc-fdroid.cfg").is_file())
        self.assertFalse(Path("NONE").exists())

    def test_setup_a_new_repo_with_no_keystore_add_apk_and_update(self):
        Path("fdroid-icon.png").touch()
        Path("repo").mkdir()
        shutil.copy(FILES / "urzip.apk", "repo")
        # this should fail because this repo has no keystore
        self.assert_run_fail(
            self.fdroid_cmd + ["update", "--create-metadata", "--verbose"]
        )

        # now set up fake, non-working keystore setup
        Path("keystore.p12").touch()
        self.update_yaml(
            common.CONFIG_FILE,
            {
                "keystore": "keystore.p12",
                "repo_keyalias": "foo",
                "keystorepass": "foo",
                "keypass": "foo",
            },
        )
        # this should fail because this repo has a bad/fake keystore
        self.assert_run_fail(
            self.fdroid_cmd + ["update", "--create-metadata", "--verbose"]
        )

    def test_copy_tests_repo_update_with_binary_transparency_log(self):
        self.fdroid_init_with_prebuilt_keystore()
        shutil.copytree(FILES / "repo", "repo", dirs_exist_ok=True)
        shutil.copytree(FILES / "metadata", "metadata")
        git_remote = self.testdir / "git_remote"
        self.update_yaml(
            common.CONFIG_FILE, {"binary_transparency_remote": str(git_remote)}
        )
        self.assert_run(self.fdroid_cmd + ["update", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["deploy", "--verbose"])
        self.assertIn("<application id=", Path("repo/index.xml").read_text())
        self.assertTrue(Path("repo/index.jar").is_file())
        self.assertTrue(Path("repo/index-v1.jar").is_file())
        os.chdir("binary_transparency")
        proc = self.assert_run(
            ["git", "rev-list", "--count", "HEAD"], capture_output=True
        )
        self.assertEqual(int(proc.stdout), 2)
        os.chdir(git_remote)
        proc = self.assert_run(
            ["git", "rev-list", "--count", "HEAD"], capture_output=True
        )
        self.assertEqual(int(proc.stdout), 2)

    def test_setup_a_new_repo_with_keystore_with_apk_update_then_without_key(self):
        shutil.copy(FILES / "keystore.jks", "keystore.jks")
        self.fdroid_init_with_prebuilt_keystore("keystore.jks")
        shutil.copy(FILES / "urzip.apk", "repo")
        self.assert_run(self.fdroid_cmd + ["update", "--create-metadata", "--verbose"])
        self.assert_run(self.fdroid_cmd + ["readmeta"])
        self.assertIn("<application id=", Path("repo/index.xml").read_text())
        self.assertTrue(Path("repo/index.jar").is_file())
        self.assertTrue(Path("repo/index-v1.jar").is_file())
        apkcache = Path("tmp/apkcache.json")
        self.assertTrue(apkcache.is_file())
        self.assertTrue(apkcache.stat().st_size > 0)

        # now set fake repo_keyalias
        self.update_yaml(common.CONFIG_FILE, {"repo_keyalias": "fake"})
        # this should fail because this repo has a bad repo_keyalias
        self.assert_run_fail(self.fdroid_cmd + ["update"])

        # this should fail because a keystore is already there
        self.assert_run_fail(self.fdroid_cmd + ["update", "--create-key"])

        # now actually create the key with the existing settings
        Path("keystore.jks").unlink()
        self.assert_run(self.fdroid_cmd + ["update", "--create-key"])
        self.assertTrue(Path("keystore.jks").is_file())

    def test_setup_a_new_repo_from_scratch_using_android_home_env_var_with_git_mirror(
        self,
    ):
        server_git_mirror = self.testdir / "server_git_mirror"
        server_git_mirror.mkdir()
        self.assert_run(
            ["git", "-C", server_git_mirror, "init", "--initial-branch", "master"]
        )
        self.assert_run(
            [
                "git",
                "-C",
                server_git_mirror,
                "config",
                "receive.denyCurrentBranch",
                "updateInstead",
            ]
        )

        self.fdroid_init_with_prebuilt_keystore()
        self.update_yaml(
            common.CONFIG_FILE,
            {"archive_older": 3, "servergitmirrors": str(server_git_mirror)},
        )
        for f in FILES.glob("repo/com.politedroid_[345].apk"):
            shutil.copy(f, "repo")
        self.assert_run(self.fdroid_cmd + ["update", "--create-metadata"])
        self.assert_run(self.fdroid_cmd + ["deploy"])
        git_mirror = Path("git-mirror")
        self.assertTrue((git_mirror / "fdroid/repo/com.politedroid_3.apk").is_file())
        self.assertTrue((git_mirror / "fdroid/repo/com.politedroid_4.apk").is_file())
        self.assertTrue((git_mirror / "fdroid/repo/com.politedroid_5.apk").is_file())
        self.assertTrue(
            (server_git_mirror / "fdroid/repo/com.politedroid_3.apk").is_file()
        )
        self.assertTrue(
            (server_git_mirror / "fdroid/repo/com.politedroid_4.apk").is_file()
        )
        self.assertTrue(
            (server_git_mirror / "fdroid/repo/com.politedroid_5.apk").is_file()
        )
        (git_mirror / ".git/test-stamp").write_text(str(datetime.now()))

        # add one more APK to trigger archiving
        shutil.copy(FILES / "repo/com.politedroid_6.apk", "repo")
        self.assert_run(self.fdroid_cmd + ["update"])
        self.assert_run(self.fdroid_cmd + ["deploy"])
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertFalse((git_mirror / "fdroid/archive/com.politedroid_3.apk").exists())
        self.assertFalse(
            (server_git_mirror / "fdroid/archive/com.politedroid_3.apk").exists()
        )
        self.assertTrue((git_mirror / "fdroid/repo/com.politedroid_4.apk").is_file())
        self.assertTrue((git_mirror / "fdroid/repo/com.politedroid_5.apk").is_file())
        self.assertTrue((git_mirror / "fdroid/repo/com.politedroid_6.apk").is_file())
        self.assertTrue(
            (server_git_mirror / "fdroid/repo/com.politedroid_4.apk").is_file()
        )
        self.assertTrue(
            (server_git_mirror / "fdroid/repo/com.politedroid_5.apk").is_file()
        )
        self.assertTrue(
            (server_git_mirror / "fdroid/repo/com.politedroid_6.apk").is_file()
        )
        before = sum(
            f.stat().st_size for f in (git_mirror / ".git").glob("**/*") if f.is_file()
        )

        self.update_yaml(common.CONFIG_FILE, {"git_mirror_size_limit": "60kb"})
        self.assert_run(self.fdroid_cmd + ["update"])
        self.assert_run(self.fdroid_cmd + ["deploy"])
        self.assertTrue(Path("archive/com.politedroid_3.apk").is_file())
        self.assertFalse(
            (server_git_mirror / "fdroid/archive/com.politedroid_3.apk").exists()
        )
        after = sum(
            f.stat().st_size for f in (git_mirror / ".git").glob("**/*") if f.is_file()
        )
        self.assertFalse((git_mirror / ".git/test-stamp").exists())
        self.assert_run(["git", "-C", git_mirror, "gc"])
        self.assert_run(["git", "-C", server_git_mirror, "gc"])
        self.assertGreater(before, after)

    def test_sign_binary_repo_in_offline_box_then_publishing_from_online_box(self):
        offline_root = self.testdir / "offline_root"
        offline_root.mkdir()
        local_copy_dir = self.testdir / "local_copy_dir/fdroid"
        local_copy_dir.mkdir(parents=True)
        online_root = self.testdir / "online_root"
        online_root.mkdir()
        server_web_root = self.testdir / "server_web_root/fdroid"
        server_web_root.mkdir(parents=True)

        # create offline binary transparency log
        (offline_root / "binary_transparency").mkdir()
        os.chdir(offline_root / "binary_transparency")
        self.assert_run(["git", "init", "--initial-branch", "master"])

        # fake git remote server for binary transparency log
        binary_transparency_remote = self.testdir / "binary_transparency_remote"
        binary_transparency_remote.mkdir()

        # fake git remote server for repo mirror
        server_git_mirror = self.testdir / "server_git_mirror"
        server_git_mirror.mkdir()
        os.chdir(server_git_mirror)
        self.assert_run(["git", "init", "--initial-branch", "master"])
        self.assert_run(["git", "config", "receive.denyCurrentBranch", "updateInstead"])

        os.chdir(offline_root)
        self.fdroid_init_with_prebuilt_keystore()
        shutil.copytree(FILES / "repo", "repo", dirs_exist_ok=True)
        shutil.copytree(FILES / "metadata", "metadata")
        Path("unsigned").mkdir()
        shutil.copy(FILES / "urzip-release-unsigned.apk", "unsigned")
        self.update_yaml(
            common.CONFIG_FILE,
            {
                "archive_older": 3,
                "mirrors": [
                    "http://foo.bar/fdroid",
                    "http://asdflkdsfjafdsdfhkjh.onion/fdroid",
                ],
                "servergitmirrors": str(server_git_mirror),
                "local_copy_dir": str(local_copy_dir),
            },
        )
        self.assert_run(self.fdroid_cmd + ["update", "--pretty"])
        index_xml = Path("repo/index.xml").read_text()
        self.assertIn("<application id=", index_xml)
        self.assertIn("/fdroid/repo</mirror>", index_xml)
        mirror_cnt = sum(1 for line in index_xml.splitlines() if "<mirror>" in line)
        self.assertEqual(mirror_cnt, 2)

        archive_xml = Path("archive/index.xml").read_text()
        self.assertIn("/fdroid/archive</mirror>", archive_xml)
        mirror_cnt = sum(1 for line in archive_xml.splitlines() if "<mirror>" in line)
        self.assertEqual(mirror_cnt, 2)

        os.chdir("binary_transparency")
        proc = self.assert_run(
            ["git", "rev-list", "--count", "HEAD"], capture_output=True
        )
        self.assertEqual(int(proc.stdout), 1)
        os.chdir(offline_root)
        self.assert_run(self.fdroid_cmd + ["deploy", "--verbose"])
        self.assertTrue(
            Path(local_copy_dir / "unsigned/urzip-release-unsigned.apk").is_file()
        )
        self.assertIn(
            "<application id=", (local_copy_dir / "repo/index.xml").read_text()
        )
        os.chdir(online_root)
        self.update_yaml(
            common.CONFIG_FILE,
            {
                "local_copy_dir": str(local_copy_dir),
                "sync_from_local_copy_dir": True,
                "serverwebroot": str(server_web_root),
                "servergitmirrors": str(server_git_mirror),
                "binary_transparency_remote": str(binary_transparency_remote),
            },
        )
        self.assert_run(self.fdroid_cmd + ["deploy", "--verbose"])
        self.assertTrue((online_root / "unsigned/urzip-release-unsigned.apk").is_file())
        self.assertTrue(
            (server_web_root / "unsigned/urzip-release-unsigned.apk").is_file()
        )
        os.chdir(binary_transparency_remote)
        proc = self.assert_run(
            ["git", "rev-list", "--count", "HEAD"], capture_output=True
        )
        self.assertEqual(int(proc.stdout), 1)

        if platform.system() == 'Darwin':
            self.skipTest('FIXME the last step of this test fails only on macOS')
        os.chdir(server_git_mirror)
        proc = self.assert_run(
            ["git", "rev-list", "--count", "HEAD"], capture_output=True
        )
        self.assertEqual(int(proc.stdout), 1)

    @unittest.skipUnless(USE_APKSIGNER, "requires apksigner")
    def test_extracting_and_publishing_with_developer_signature(self):
        self.fdroid_init_with_prebuilt_keystore()
        self.update_yaml(
            common.CONFIG_FILE,
            {
                "keydname": "CN=Birdman, OU=Cell, O=Alcatraz, L=Alcatraz, S=California, C=US"
            },
        )
        Path("metadata").mkdir()
        shutil.copy(FILES / "metadata/com.politedroid.yml", "metadata")
        Path("unsigned").mkdir()
        shutil.copy(FILES / "repo/com.politedroid_6.apk", "unsigned")
        self.assert_run(
            self.fdroid_cmd + ["signatures", "unsigned/com.politedroid_6.apk"]
        )
        self.assertTrue(
            Path("metadata/com.politedroid/signatures/6/MANIFEST.MF").is_file()
        )
        self.assertTrue(
            Path("metadata/com.politedroid/signatures/6/RELEASE.RSA").is_file()
        )
        self.assertTrue(
            Path("metadata/com.politedroid/signatures/6/RELEASE.SF").is_file()
        )
        self.assertFalse(Path("repo/com.politedroid_6.apk").exists())
        self.assert_run(self.fdroid_cmd + ["publish"])
        self.assertTrue(Path("repo/com.politedroid_6.apk").is_file())
        if shutil.which("apksigner"):
            self.assert_run(["apksigner", "verify", "repo/com.politedroid_6.apk"])
        if shutil.which("jarsigner"):
            self.assert_run(["jarsigner", "-verify", "repo/com.politedroid_6.apk"])

    @unittest.skipUnless(shutil.which("wget"), "requires wget")
    def test_mirroring_a_repo(self):
        """Start a local webserver to mirror a fake repo from.

        Proxy settings via environment variables can interfere with
        this test. The requests library will automatically pick up
        proxy settings from environment variables. Proxy settings can
        force the local connection over the proxy, which might not
        support that, then this fails with an error like 405 or
        others.

        """
        tmp_test = self.testdir / "tests"
        tmp_test.mkdir()
        shutil.copytree(FILES, tmp_test, dirs_exist_ok=True)
        os.chdir(tmp_test)
        Path("archive").mkdir(exist_ok=True)
        shutil.copy("repo/index-v1.json", self.tmp_repo_root)
        self.assert_run(self.fdroid_cmd + ["update"])
        self.assert_run(self.fdroid_cmd + ["signindex"])
        shutil.move(self.tmp_repo_root / "index-v1.json", "repo/index-v1.json")

        class RequestHandler(SimpleHTTPRequestHandler):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, directory=tmp_test, **kwargs)

        httpd = ThreadingHTTPServer(("127.0.0.1", 0), RequestHandler)
        threading.Thread(target=httpd.serve_forever).start()

        os.chdir(self.tmp_repo_root)
        host, port = httpd.socket.getsockname()
        url, output_dir = f"http://{host}:{port}/", Path(f"{host}:{port}")
        self.assert_run(self.fdroid_cmd + ["mirror", url])
        self.assertTrue((output_dir / "repo/souch.smsbypass_9.apk").is_file())
        self.assertTrue((output_dir / "repo/icons-640/souch.smsbypass.9.png").is_file())
        # the index shouldn't be saved unless it was verified
        self.assertFalse((output_dir / "repo/index-v1.jar").exists())
        self.assert_run_fail(
            self.fdroid_cmd + ["mirror", f"{url}?fingerprint=asdfasdf"]
        )
        self.assertFalse((output_dir / "repo/index-v1.jar").exists())
        self.assert_run(
            self.fdroid_cmd
            + [
                "mirror",
                f"{url}?fingerprint=F49AF3F11EFDDF20DFFD70F5E3117B9976674167ADCA280E6B1932A0601B26F6",
            ],
        )
        self.assertTrue((output_dir / "repo/index-v1.jar").is_file())

        httpd.shutdown()

    def test_recovering_from_broken_git_submodules(self):
        Path("foo").mkdir()
        Path("bar").mkdir()
        os.chdir("foo")
        self.assert_run(["git", "init"])
        Path("a").write_text("a")
        self.assert_run(["git", "add", "a"])
        self.assert_run(["git", "commit", "-m", "a"])

        os.chdir("../bar")
        self.assert_run(["git", "init"])
        self.assert_run(
            ["git", "submodule", "add", f"file://{Path().resolve()}/../foo", "baz"]
        )
        Path(".gitmodules").unlink()
        self.assert_run(["git", "commit", "-am", "a"])
        shutil.rmtree("baz")
        self.assert_run(["git", "checkout", "baz"])
        self.assert_run(["git", "tag", "2"])

        os.chdir("..")
        Path("repo").mkdir()
        Path("metadata").mkdir()
        self.update_yaml(
            "metadata/fake.yml",
            {
                "RepoType": "git",
                "Repo": f"file://{Path().resolve()}/bar",
                "AutoUpdateMode": "Version",
                "UpdateCheckMode": "Tags",
                "UpdateCheckData": "|||",
                "CurrentVersion": 1,
                "CurrentVersionCode": 1,
            },
        )
        self.assert_run(self.fdroid_cmd + ["checkupdates", "--allow-dirty"])
        self.assertIn("CurrentVersionCode: 2", Path("metadata/fake.yml").read_text())

    def test_checkupdates_ignore_broken_submodule(self):
        Path("foo").mkdir()
        Path("bar").mkdir()
        os.chdir("foo")
        self.assert_run(["git", "init"])
        Path("a").write_text("a")
        self.assert_run(["git", "add", "a"])
        self.assert_run(["git", "commit", "-m", "a"])

        os.chdir("../bar")
        self.assert_run(["git", "init"])
        self.assert_run(
            ["git", "submodule", "add", f"file://{Path().resolve()}/../foo", "baz"]
        )
        self.assert_run(["git", "commit", "-am", "a"])
        self.assert_run(["git", "tag", "2"])

        os.chdir("../foo")
        # delete the commit referenced in bar
        self.assert_run(["git", "commit", "--amend", "-m", "aa"])
        self.assert_run(["git", "reflog", "expire", "--expire", "now", "--all"])
        self.assert_run(["git", "gc", "--aggressive", "--prune=now"])

        os.chdir("..")

        Path("repo").mkdir()
        Path("metadata").mkdir()
        self.update_yaml(
            "metadata/fake.yml",
            {
                "RepoType": "git",
                "Repo": f"file://{Path().resolve()}/bar",
                "Builds": [{"versionName": 1, "versionCode": 1, "submodules": True}],
                "AutoUpdateMode": "Version",
                "UpdateCheckMode": "Tags",
                "UpdateCheckData": "|||",
                "CurrentVersion": 1,
                "CurrentVersionCode": 1,
            },
        )
        self.assert_run(self.fdroid_cmd + ["checkupdates", "--allow-dirty"])
        self.assertIn("CurrentVersionCode: 2", Path("metadata/fake.yml").read_text())

    def test_checkupdates_check_version_in_submodule(self):
        Path("app").mkdir()
        Path("sub").mkdir()
        os.chdir("sub")
        self.assert_run(["git", "init"])
        Path("ver").write_text("1")
        self.assert_run(["git", "add", "ver"])
        self.assert_run(["git", "commit", "-m", "1"])

        os.chdir("../app")
        self.assert_run(["git", "init"])
        self.assert_run(
            ["git", "submodule", "add", f"file://{Path().resolve()}/../sub"]
        )
        self.assert_run(["git", "commit", "-am", "1"])
        self.assert_run(["git", "tag", "1"])

        os.chdir("../sub")
        Path("ver").write_text("2")
        self.assert_run(["git", "commit", "-am", "2"])

        os.chdir("../app")
        self.assert_run(["git", "init"])
        self.assert_run(["git", "submodule", "update", "--remote"])
        self.assert_run(["git", "commit", "-am", "2"])

        os.chdir("..")
        Path("repo").mkdir()
        Path("metadata").mkdir()
        self.update_yaml(
            "metadata/fake.yml",
            {
                "RepoType": "git",
                "Repo": f"file://{Path().resolve()}/app",
                "Builds": [{"versionName": 0, "versionCode": 0, "submodules": True}],
                "AutoUpdateMode": "Version",
                "UpdateCheckMode": "Tags",
                "UpdateCheckData": r"sub/ver|(\d)||",
                "CurrentVersion": 0,
                "CurrentVersionCode": 0,
            },
        )
        self.assert_run(
            self.fdroid_cmd + ["checkupdates", "--allow-dirty", "--auto", "-v"]
        )
        self.assertIn("CurrentVersionCode: 1", Path("metadata/fake.yml").read_text())



#!/usr/bin/env python3

import os
import fdroidserver
import shutil
import sys
from fdroidserver import common, nightly

if os.getenv('CI') is None:
    print('ERROR: This can overwrite SSH keys, so it should only be run in CI')
    sys.exit(1)

os.chdir(os.path.dirname(__file__))
config = fdroidserver.common.read_config()
nightly.PASSWORD = config['keystorepass']
nightly.KEY_ALIAS = config['repo_keyalias']

privkey = nightly._ssh_key_from_debug_keystore('keystore.jks')
print('privkey', privkey)
ssh_private_key_file = os.path.join(os.getenv('HOME'), '.ssh', 'id_rsa')
if os.path.exists(ssh_private_key_file):
    print('ERROR:', ssh_private_key_file, 'exists!')
    sys.exit(1)
shutil.move(privkey, ssh_private_key_file)
shutil.move(privkey + '.pub', ssh_private_key_file + '.pub')



#!/usr/bin/env python3

import os
import pkgutil
import textwrap
import unittest
import tempfile
from unittest import mock

import fdroidserver.__main__
from .shared_test_code import TmpCwd, TmpPyPath


class MainTest(unittest.TestCase):
    '''this tests fdroid.py'''

    def test_COMMANDS_check(self):
        """make sure the built in sub-command defs didn't change unintentionally"""
        self.assertListEqual(
            [x for x in fdroidserver.__main__.COMMANDS],
            [
                'build',
                'init',
                'publish',
                'gpgsign',
                'update',
                'deploy',
                'verify',
                'checkupdates',
                'import',
                'install',
                'readmeta',
                'rewritemeta',
                'lint',
                'scanner',
                'signindex',
                'btlog',
                'signatures',
                'nightly',
                'mirror',
            ],
        )

    def test_call_init(self):
        co = mock.Mock()
        with mock.patch('sys.argv', ['', 'init', '-h']):
            with mock.patch('fdroidserver.init.main', co):
                with mock.patch('sys.exit') as exit_mock:
                    fdroidserver.__main__.main()
                    # note: this is sloppy, if `init` changes
                    # this might need changing too
                    exit_mock.assert_called_once_with(0)
        co.assert_called_once_with()

    def test_call_deploy(self):
        co = mock.Mock()
        with mock.patch('sys.argv', ['', 'deploy', '-h']):
            with mock.patch('fdroidserver.deploy.main', co):
                with mock.patch('sys.exit') as exit_mock:
                    fdroidserver.__main__.main()
                    # note: this is sloppy, if `deploy` changes
                    # this might need changing too
                    exit_mock.assert_called_once_with(0)
        co.assert_called_once_with()

    def test_find_plugins(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with open('fdroid_testy1.py', 'w') as f:
                f.write(
                    textwrap.dedent(
                        """\
                        fdroid_summary = "ttt"
                        main = lambda: 'all good'"""
                    )
                )
            with TmpPyPath(tmpdir):
                plugins = fdroidserver.__main__.find_plugins()
                self.assertIn('testy1', plugins.keys())
                self.assertEqual(plugins['testy1']['summary'], 'ttt')
                self.assertEqual(
                    __import__(
                        plugins['testy1']['name'], None, None, ['testy1']
                    ).main(),
                    'all good',
                )

    def test_main_plugin_lambda(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with open('fdroid_testy2.py', 'w') as f:
                f.write(
                    textwrap.dedent(
                        """\
                        fdroid_summary = "ttt"
                        main = lambda: print('all good')"""
                    )
                )
            with TmpPyPath(tmpdir):
                with mock.patch('sys.argv', ['', 'testy2']):
                    with mock.patch('sys.exit') as exit_mock:
                        fdroidserver.__main__.main()
                        exit_mock.assert_called_once_with(0)

    def test_main_plugin_def(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with open('fdroid_testy3.py', 'w') as f:
                f.write(
                    textwrap.dedent(
                        """\
                        fdroid_summary = "ttt"
                        def main():
                            print('all good')"""
                    )
                )
            with TmpPyPath(tmpdir):
                with mock.patch('sys.argv', ['', 'testy3']):
                    with mock.patch('sys.exit') as exit_mock:
                        fdroidserver.__main__.main()
                        exit_mock.assert_called_once_with(0)

    def test_main_broken_plugin(self):
        """making sure broken plugins get their exceptions through"""
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with open('fdroid_testy4.py', 'w') as f:
                f.write(
                    textwrap.dedent(
                        """\
                        fdroid_summary = "ttt"
                        def main():
                            raise Exception("this plugin is broken")"""
                    )
                )
            with TmpPyPath(tmpdir):
                with mock.patch('sys.argv', ['', 'testy4']):
                    with self.assertRaisesRegex(Exception, "this plugin is broken"):
                        fdroidserver.__main__.main()

    def test_main_malicious_plugin(self):
        """The purpose of this test is to make sure code in plugins
        doesn't get executed unintentionally.
        """
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with open('fdroid_testy5.py', 'w') as f:
                f.write(
                    textwrap.dedent(
                        """\
                        fdroid_summary = "ttt"
                        raise Exception("this plugin is malicious")
                        def main():
                            print("evil things")"""
                    )
                )
            with TmpPyPath(tmpdir):
                with mock.patch('sys.argv', ['', 'lint']):
                    with mock.patch('sys.exit') as exit_mock:
                        fdroidserver.__main__.main()
                        # note: this is sloppy, if `lint` changes
                        # this might need changing too
                        exit_mock.assert_called_once_with(0)

    def test_main_prevent_plugin_override(self):
        """making sure build-in subcommands cannot be overridden by plugins"""
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with open('fdroid_signatures.py', 'w') as f:
                f.write(
                    textwrap.dedent(
                        """\
                        fdroid_summary = "ttt"
                        def main():
                            raise("plugin overrides don't get prevent!")"""
                    )
                )
            with TmpPyPath(tmpdir):
                with mock.patch('sys.argv', ['', 'signatures']):
                    with mock.patch('sys.exit') as exit_mock:
                        fdroidserver.__main__.main()
                        # note: this is sloppy, if `signatures` changes
                        # this might need changing too
                        self.assertEqual(exit_mock.call_count, 2)

    def test_preparse_plugin_lookup_bad_name(self):
        self.assertRaises(
            ValueError,
            fdroidserver.__main__.preparse_plugin,
            "some.package",
            "/non/existent/module/path",
        )

    def test_preparse_plugin_lookup_bad_path(self):
        self.assertRaises(
            ValueError,
            fdroidserver.__main__.preparse_plugin,
            "fake_module_name",
            "/non/existent/module/path",
        )

    def test_preparse_plugin_lookup_summary_missing(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with open('fdroid_testy6.py', 'w') as f:
                f.write("main = lambda: print('all good')")
            with TmpPyPath(tmpdir):
                p = [x for x in pkgutil.iter_modules() if x[1].startswith('fdroid_')]
                module_dir = p[0][0].path
                module_name = p[0][1]
                self.assertRaises(
                    NameError,
                    fdroidserver.__main__.preparse_plugin,
                    module_name,
                    module_dir,
                )

    def test_preparse_plugin_lookup_module_file(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with open('fdroid_testy7.py', 'w') as f:
                f.write(
                    textwrap.dedent(
                        """\
                        fdroid_summary = "ttt"
                        main = lambda: pritn('all good')"""
                    )
                )
            with TmpPyPath(tmpdir):
                p = [x for x in pkgutil.iter_modules() if x[1].startswith('fdroid_')]
                module_path = p[0][0].path
                module_name = p[0][1]
                d = fdroidserver.__main__.preparse_plugin(module_name, module_path)
            self.assertDictEqual(d, {'name': 'fdroid_testy7', 'summary': 'ttt'})

    def test_preparse_plugin_lookup_module_dir(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            os.mkdir(os.path.join(tmpdir, 'fdroid_testy8'))
            with open('fdroid_testy8/__main__.py', 'w') as f:
                f.write(
                    textwrap.dedent(
                        """\
                        fdroid_summary = "ttt"
                        main = lambda: print('all good')"""
                    )
                )
            with open('fdroid_testy8/__init__.py', 'w') as f:
                pass
            with TmpPyPath(tmpdir):
                p = [x for x in pkgutil.iter_modules() if x[1].startswith('fdroid_')]
                module_path = p[0][0].path
                module_name = p[0][1]
                d = fdroidserver.__main__.preparse_plugin(module_name, module_path)
            self.assertDictEqual(d, {'name': 'fdroid_testy8', 'summary': 'ttt'})



#!/usr/bin/env python3

import os
import textwrap
import unittest

from pathlib import Path
from unittest.mock import Mock, patch

import fdroidserver
from fdroidserver import common, install
from fdroidserver.exception import BuildException, FDroidException


@unittest.skipIf(os.uname().machine == 's390x', 'adb is not ported to s390x')
class InstallTest(unittest.TestCase):
    '''fdroidserver/install.py'''

    def tearDown(self):
        common.config = None

    def test_devices(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        try:
            config['adb'] = fdroidserver.common.find_sdk_tools_cmd('adb')
        except FDroidException as e:
            self.skipTest(f'Skipping test because: {e}')
        self.assertTrue(os.path.exists(config['adb']))
        self.assertTrue(os.path.isfile(config['adb']))
        devices = fdroidserver.install.devices()
        self.assertIsInstance(devices, list, 'install.devices() did not return a list!')
        for device in devices:
            self.assertIsInstance(device, str)

    def test_devices_fail(self):
        common.config = dict()
        common.fill_config_defaults(common.config)
        common.config['adb'] = '/bin/false'
        with self.assertRaises(FDroidException):
            fdroidserver.install.devices()

    def test_devices_fail_nonexistent(self):
        """This is mostly just to document this strange difference in behavior"""
        common.config = dict()
        common.fill_config_defaults(common.config)
        common.config['adb'] = '/nonexistent'
        with self.assertRaises(BuildException):
            fdroidserver.install.devices()

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_none(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = 'List of devices attached\n\n'
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_one(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = 'List of devices attached\n05995813\tdevice\n\n'
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual(['05995813'], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_many(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """* daemon not running; starting now at tcp:5037
            * daemon started successfully
            List of devices attached
            RZCT809FTQM	device
            05995813	device
            emulator-5556	device
            emulator-5554	unauthorized
            0a388e93	no permissions (missing udev rules? user is in the plugdev group); see [http://developer.android.com/tools/device.html]
            986AY133QL	device
            09301JEC215064	device
            015d165c3010200e	device
            4DCESKVGUC85VOTO	device

            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual(
            [
                'RZCT809FTQM',
                '05995813',
                'emulator-5556',
                '986AY133QL',
                '09301JEC215064',
                '015d165c3010200e',
                '4DCESKVGUC85VOTO',
            ],
            fdroidserver.install.devices(),
        )

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_error(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """* daemon not running. starting it now on port 5037 *
            * daemon started successfully *
            ** daemon still not running
            error: cannot connect to daemon
            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_no_permissions(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """List of devices attached
            ????????????????	no permissions
            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_unauthorized(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """List of devices attached
            aeef5e4e	unauthorized
            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_no_permissions_with_serial(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """List of devices attached
             4DCESKVGUC85VOTO	no permissions (missing udev rules? user is in the plugdev group); see [http://developer.android.com/tools/device.html]

            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @staticmethod
    def _download_raise(privacy_mode):
        raise Exception('fake failed download')

    @patch('fdroidserver.install.download_apk')
    @patch('fdroidserver.install.download_fdroid_apk')
    @patch('fdroidserver.install.download_fdroid_apk_from_github')
    @patch('fdroidserver.install.download_fdroid_apk_from_ipns')
    @patch('fdroidserver.install.download_fdroid_apk_from_maven')
    def test_install_fdroid_apk_privacy_mode_true(
        self, maven, ipns, github, download_fdroid_apk, download_apk
    ):
        download_apk.side_effect = self._download_raise
        download_fdroid_apk.side_effect = self._download_raise
        github.side_effect = self._download_raise
        ipns.side_effect = self._download_raise
        maven.side_effect = self._download_raise
        fdroidserver.common.config = {'jarsigner': 'fakepath'}
        install.install_fdroid_apk(privacy_mode=True)
        download_apk.assert_not_called()
        download_fdroid_apk.assert_not_called()
        github.assert_called_once()
        ipns.assert_called_once()
        maven.assert_called_once()

    @patch('fdroidserver.install.download_apk')
    @patch('fdroidserver.install.download_fdroid_apk')
    @patch('fdroidserver.install.download_fdroid_apk_from_github')
    @patch('fdroidserver.install.download_fdroid_apk_from_ipns')
    @patch('fdroidserver.install.download_fdroid_apk_from_maven')
    def test_install_fdroid_apk_privacy_mode_false(
        self, maven, ipns, github, download_fdroid_apk, download_apk
    ):
        download_apk.side_effect = self._download_raise
        download_fdroid_apk.side_effect = self._download_raise
        github.side_effect = self._download_raise
        ipns.side_effect = self._download_raise
        maven.side_effect = self._download_raise
        fdroidserver.common.config = {'jarsigner': 'fakepath'}
        install.install_fdroid_apk(privacy_mode=False)
        download_apk.assert_called_once()
        download_fdroid_apk.assert_called_once()
        github.assert_called_once()
        ipns.assert_called_once()
        maven.assert_called_once()

    @patch('fdroidserver.install.download_apk')
    @patch('fdroidserver.install.download_fdroid_apk')
    @patch('fdroidserver.install.download_fdroid_apk_from_github')
    @patch('fdroidserver.install.download_fdroid_apk_from_ipns')
    @patch('fdroidserver.install.download_fdroid_apk_from_maven')
    @patch('locale.getlocale', lambda: ('zh_CN', 'UTF-8'))
    def test_install_fdroid_apk_privacy_mode_locale_auto(
        self, maven, ipns, github, download_fdroid_apk, download_apk
    ):
        download_apk.side_effect = self._download_raise
        download_fdroid_apk.side_effect = self._download_raise
        github.side_effect = self._download_raise
        ipns.side_effect = self._download_raise
        maven.side_effect = self._download_raise
        fdroidserver.common.config = {'jarsigner': 'fakepath'}
        install.install_fdroid_apk(privacy_mode=None)
        download_apk.assert_not_called()
        download_fdroid_apk.assert_not_called()
        github.assert_called_once()
        ipns.assert_called_once()
        maven.assert_called_once()

    @patch('fdroidserver.net.download_using_mirrors', lambda m: 'testvalue')
    def test_download_fdroid_apk_smokecheck(self):
        self.assertEqual('testvalue', install.download_fdroid_apk())

    @unittest.skipUnless(os.getenv('test_download_fdroid_apk'), 'requires net access')
    def test_download_fdroid_apk(self):
        f = install.download_fdroid_apk()
        self.assertTrue(Path(f).exists())

    @unittest.skipUnless(os.getenv('test_download_fdroid_apk'), 'requires net access')
    def test_download_fdroid_apk_from_maven(self):
        f = install.download_fdroid_apk_from_maven()
        self.assertTrue(Path(f).exists())

    @unittest.skipUnless(os.getenv('test_download_fdroid_apk'), 'requires net access')
    def test_download_fdroid_apk_from_ipns(self):
        f = install.download_fdroid_apk_from_ipns()
        self.assertTrue(Path(f).exists())

    @unittest.skipUnless(os.getenv('test_download_fdroid_apk'), 'requires net access')
    def test_download_fdroid_apk_from_github(self):
        f = install.download_fdroid_apk_from_github()
        self.assertTrue(Path(f).exists())
